"""Testing for the helpers module."""

import re
from collections import defaultdict
from datetime import datetime
from pathlib import Path
from unittest import mock

import pytest

from djtools.collection.base_collection import Collection
from djtools.collection.base_playlist import Playlist
from djtools.collection.base_track import Track
from djtools.collection.config import (
    PlaylistConfigContent,
    PlaylistName,
    RegisteredPlatforms,
)
from djtools.collection.helpers import (
    add_selectors_to_tags,
    aggregate_playlists,
    BooleanNode,
    build_combiner_playlists,
    build_tag_playlists,
    copy_file,
    DATE_SELECTOR_REGEX,
    filter_tag_playlists,
    INEQUALITY_MAP,
    parse_expression,
    parse_numerical_selectors,
    parse_string_selectors,
    parse_timedelta,
    print_data,
    print_playlists_tag_statistics,
    scale_data,
)
from djtools.collection.platform_registry import PLATFORM_REGISTRY
from djtools.collection.rekordbox_collection import RekordboxCollection
from djtools.collection.rekordbox_playlist import RekordboxPlaylist


# pylint: disable=duplicate-code


def test_copy_file(tmpdir, rekordbox_track):
    """Test for the copy_file function."""
    dest_dir = Path(tmpdir) / "output"
    dest_dir.mkdir(parents=True, exist_ok=True)
    file_name = rekordbox_track.get_location().name
    copy_file(track=rekordbox_track, destination=dest_dir)
    new_file_path = dest_dir / file_name
    # NOTE(a-rich): `Location` attributes in the XML's `TRACK` tags always
    # have unix-style paths so comparisons made with paths created in Windows
    # must be interpreted `.as_posix()`.
    assert rekordbox_track.get_location() == new_file_path
    assert new_file_path.exists()


def test_platform_registry_structure():
    """Test for the PLATFORM_REGISTRY object."""
    assert isinstance(PLATFORM_REGISTRY, dict)
    assert len(PLATFORM_REGISTRY) > 0
    required_class_impl_keys = {"collection", "playlist", "track"}
    required_base_class_impls = {Collection, Playlist, Track}
    for registered_software, impls in PLATFORM_REGISTRY.items():
        assert isinstance(registered_software, RegisteredPlatforms)
        assert isinstance(impls, dict)
        assert set(impls.keys()) == required_class_impl_keys
        assert (
            set(
                base
                for class_impl in impls.values()
                for base in class_impl.__bases__
            )
            == required_base_class_impls
        )


def test_build_tag_playlists_minimum_tracks_config():
    """Test the build_tag_playlists function."""
    playlist_content = PlaylistConfigContent(
        name="playlists",
        playlists=["Tag"],
    )
    playlists = build_tag_playlists(
        playlist_content,
        {"Tag": {1: None}},
        RekordboxPlaylist,
        minimum_tracks=2,
    )
    assert playlists is None


@mock.patch(
    "djtools.collection.rekordbox_track.RekordboxTrack.get_genre_tags",
    mock.Mock(return_value=["Tag"]),
)
def test_build_tag_playlists_pure_playlist_minimum_tracks_config(
    rekordbox_collection,
):
    """Test the build_tag_playlists function."""
    tracks = rekordbox_collection.get_tracks()
    example_track = tracks[list(tracks)[0]]
    playlist_content = PlaylistConfigContent(
        name="playlists",
        playlists=["Pure Tag"],
    )
    playlists = build_tag_playlists(
        playlist_content,
        {"Tag": {1: example_track}},
        RekordboxPlaylist,
        minimum_tracks=2,
    )
    assert playlists is None


def test_build_tag_playlists_raises_exception_for_invalid_input():
    """Test the build_tag_playlists function."""
    with pytest.raises(
        ValueError,
        match=re.escape(f"Invalid input type {list}: {[]}"),
    ):
        build_tag_playlists([], {}, RekordboxPlaylist)


def test_build_tag_playlists_evaluates_correctly():
    """Test the build_combiner_playlists function."""
    playlist_content = PlaylistConfigContent(
        name="playlists",
        playlists=[
            "Tag",
            PlaylistName(
                tag_content="Tag",
            ),
            PlaylistConfigContent(
                name="sub-playlists",
                playlists=[
                    "Tag",
                    PlaylistName(
                        tag_content="Tag",
                        name="Inner playlist",
                    ),
                ],
            ),
        ],
    )
    playlists = build_tag_playlists(
        playlist_content, {"Tag": {1: None}}, RekordboxPlaylist
    )
    assert len(playlists.get_playlists("Tag")) == 3
    assert len(playlists.get_playlists("Inner playlist")) == 1
    assert len(playlists.get_playlists("sub-playlists")) == 1
    assert len(playlists.get_playlists("playlists")) == 1


def test_build_tag_playlists_ignores_tags(
    rekordbox_collection, playlist_config_obj
):
    """Test the build_tag_playlists function."""
    # Aggregate all the tags from the collection and isolate one tag to
    # explicitly create a playlist for; all the remaining tags will be ignored.
    tags = {
        tag
        for value in rekordbox_collection.get_all_tags().values()
        for tag in value
    }
    some_tag = next(iter(tags))
    tags.remove(some_tag)

    # Override the playlist config so that a single folder called "Some Tag" is
    # created containing a single playlist for the tag isolated as some_tag.
    # All of the remaining tags are to be ignored.
    playlist_config_obj.tags.playlists = [
        PlaylistConfigContent(
            name="_ignore",
            playlists=list(tags),
        ),
        PlaylistConfigContent(
            name="Some Tag",
            playlists=[some_tag],
        ),
    ]

    # Create a dict of tracks keyed by their individual tags.
    tags_tracks = defaultdict(dict)
    for track_id, track in rekordbox_collection.get_tracks().items():
        for tag in track.get_tags():
            tags_tracks[tag][track_id] = track

    # Build tag playlists and get the contents of it.
    playlist = build_tag_playlists(
        playlist_config_obj.tags, tags_tracks, RekordboxPlaylist, tags
    )
    tag_playlists = playlist.get_playlists()

    # build_tag_playlists doesn't build an "Unused Tags" playlist in this case
    # because we told it to ignore every tag except for some_tag.
    assert len(tag_playlists) == 1
    assert tag_playlists[0].get_name() == "Some Tag"

    # The only playlist in our "Some Tag" folder is for the tag specified in
    # some_tag.
    assert len(tag_playlists[0]) == 1
    assert tag_playlists[0][0].get_name() == some_tag


def test_build_tag_playlists_pure_playlists(
    rekordbox_collection, playlist_config_obj
):
    """Test the build_tag_playlists function."""
    playlist_config_obj.tags.playlists = ["Techno", "Pure Techno"]

    # Create a dict of tracks keyed by their individual tags.
    tags_tracks = defaultdict(dict)
    for track_id, track in rekordbox_collection.get_tracks().items():
        for tag in track.get_tags():
            tags_tracks[tag][track_id] = track

    # Build tag playlists and get the contents of it.
    playlist = build_tag_playlists(
        playlist_config_obj.tags, tags_tracks, RekordboxPlaylist
    )

    # The "Techno" playlist will contain tracks that have a "Techno" genre tag.
    techno_playlist = playlist.get_playlists("Techno")[0]
    assert len(techno_playlist) == 2
    for track in techno_playlist.get_tracks().values():
        assert "Techno" in track.get_genre_tags()

    # The "Pure Techno" playlist will only have tracks where every genre tag
    # contains the substring "Techno" (case insensitive).
    pure_techno_playlist = playlist.get_playlists("Pure Techno")[0]
    assert len(pure_techno_playlist) == 1
    for track in pure_techno_playlist.get_tracks().values():
        for tag in track.get_genre_tags():
            assert "Techno" in tag


def test_build_tag_playlists_warnings(
    rekordbox_collection, playlist_config_obj, caplog
):
    """Test the build_tag_playlists function."""
    caplog.set_level("WARNING")

    # Create a folder that will contain no playlists.
    playlists = ["test"]
    playlist_config_obj.tags.playlists = [
        PlaylistConfigContent(
            name="Folder of non-playlists",
            playlists=playlists,
        ),
    ]

    # Create a dict of tracks keyed by their individual tags.
    tags_tracks = defaultdict(dict)
    for track_id, track in rekordbox_collection.get_tracks().items():
        for tag in track.get_tags():
            tags_tracks[tag][track_id] = track

    # Build tag playlists and get the contents of it.
    playlist = build_tag_playlists(
        playlist_config_obj.tags, tags_tracks, RekordboxPlaylist
    )

    # Since "test" doesn't exist in the collection, build_tag_playlists returns
    # None. It also warns about no tracks being found with the tag "test" and
    # warns that a folder in the configuration contains no playlists.
    assert not playlist
    assert (
        caplog.records[0].message
        == f'There are no tracks with the tag "{playlists[0]}"'
    )
    assert (
        caplog.records[1].message
        == f'There were no playlists created from "{playlists}"'
    )


def test_build_combiner_playlists_minimum_tracks_config():
    """Test the build_combiner_playlists function."""
    playlist_content = PlaylistConfigContent(
        name="playlists",
        playlists=["Tag | Tag"],
    )
    playlists = build_combiner_playlists(
        playlist_content,
        {"Tag": {1: None}},
        RekordboxPlaylist,
        minimum_tracks=2,
    )
    assert len(playlists) == 0


def test_build_combiner_playlists_raises_exception_for_invalid_input():
    """Test the build_combiner_playlists function."""
    with pytest.raises(
        ValueError,
        match=re.escape(f"Invalid input type {list}: {[]}"),
    ):
        build_combiner_playlists([], {}, RekordboxPlaylist)


@mock.patch(
    "djtools.collection.helpers.parse_expression",
    side_effect=lambda x, y: {1: None},
)
def test_build_combiner_playlists_evaluates_correctly(
    mock_parse_expression,
):
    """Test the build_combiner_playlists function."""
    expected_num_playlists = 4
    playlist_content = PlaylistConfigContent(
        name="playlists",
        playlists=[
            "Tag | Tag",
            PlaylistName(
                tag_content="Tag | Tag",
            ),
            PlaylistConfigContent(
                name="sub-playlists",
                playlists=[
                    "Tag | Tag",
                    PlaylistName(
                        tag_content="Tag | Tag",
                        name="Inner playlist",
                    ),
                ],
            ),
        ],
    )
    playlists = build_combiner_playlists(
        playlist_content, {"Tag": {1: None}}, RekordboxPlaylist
    )
    assert mock_parse_expression.call_count == expected_num_playlists
    assert len(playlists.get_playlists("Tag | Tag")) == 3
    assert len(playlists.get_playlists("Inner playlist")) == 1
    assert len(playlists.get_playlists("sub-playlists")) == 1
    assert len(playlists.get_playlists("playlists")) == 1


def test_build_combiner_playlists_warnings(caplog):
    """Test the build_combiner_playlists function."""
    caplog.set_level("WARNING")

    # A valid expression must have one more operand than there are operators.
    invalid_playlist = "Invalid ~"
    content = PlaylistConfigContent(name="test", playlists=[invalid_playlist])
    build_combiner_playlists(content, {}, RekordboxPlaylist)

    # The evaluation of a BooleanNode will error out first.
    expected_error = (
        f"Error parsing expression: {invalid_playlist}\n"
        "Invalid boolean expression:\n"
        "\toperands: ['Invalid']\n"
        "\toperators: ['difference']"
    )
    assert caplog.records[0].message == expected_error

    # When evaluation of a BooleanNode fails, the result will be None rather
    # than a Playlist.
    assert caplog.records[1].message == (
        f"There are no tracks for the Combiner playlist: {invalid_playlist}"
    )

    # If none of the expressions in a folder return a playlist, then a warning
    # will be displayed.
    assert caplog.records[2].message == (
        f'There were no playlists created from "{[invalid_playlist]}"'
    )


def test_filter_tag_playlists(rekordbox_track):
    """Test the filter_tag_playlists function."""
    # This is a nested playlist with two playlists. Both contain the same
    # single track; the one one named "Filter this" will have the track removed
    # while the other will not.
    playlist = RekordboxPlaylist.new_playlist(
        "playlists",
        playlists=[
            RekordboxPlaylist.new_playlist(
                "Folder",
                playlists=[
                    RekordboxPlaylist.new_playlist(
                        "Filter this",
                        tracks={rekordbox_track.get_id(): rekordbox_track},
                    ),
                    RekordboxPlaylist.new_playlist(
                        "Some other playlist",
                        tracks={rekordbox_track.get_id(): rekordbox_track},
                    ),
                ],
            ),
            RekordboxPlaylist.new_playlist(
                "Do not filter this",
                tracks={rekordbox_track.get_id(): rekordbox_track},
            ),
            RekordboxPlaylist.new_playlist(
                "Filter this",
                tracks={rekordbox_track.get_id(): rekordbox_track},
            ),
        ],
    )

    # This is a PlaylistFilter class that will remove tracks containing the tag
    # "Gangsta" for playlists named "Filter this".
    class TestFilter:
        "PlaylistFilter implementation."

        def filter_track(self, track: Track) -> bool:
            "Filters tracks."
            return "Gangsta" not in track.get_tags()

        def is_filter_playlist(self, playlist: Playlist) -> bool:
            "Determines if playlist should have its tracks filtered."
            return playlist.get_name() == "Filter this"

    # Apply the TestFilter.
    filter_tag_playlists(playlist, [TestFilter()])

    # Playlists named "Filter this" will have no tracks.
    filtered_playlists = playlist.get_playlists("Filter this")
    for playlist in filtered_playlists:
        assert not playlist.get_tracks()

    # Other playlists will still have tracks.
    unfiltered_playlists = playlist.get_playlists("Do Not filter this")
    for playlist in unfiltered_playlists:
        assert playlist.get_tracks()


@mock.patch(
    "djtools.collection.rekordbox_playlist.RekordboxPlaylist.aggregate",
    return_value=True,
)
def test_aggregate_playlists(rekordbox_collection):
    """Test for the aggregate_playlists function."""
    # Create a playlist for each track in the collection (one track per
    # playlist).
    tracks = rekordbox_collection.get_tracks()
    playlist = RekordboxPlaylist.new_playlist(
        "Tracks",
        playlists=[
            RekordboxPlaylist.new_playlist(
                f"Track {track.get_id()}",
                tracks={track_id: track},
            )
            for track_id, track in tracks.items()
        ],
    )

    num_playlists = len(playlist)

    # Aggregate the tracks into an "All Tracks" playlist.
    aggregate_playlists(playlist, RekordboxPlaylist)

    # aggregate_playlists will have added a new playlist.
    assert len(playlist) == num_playlists + 1

    # The added playlist will be "All" plus whatever the name of the folder is.
    all_tracks_playlist = playlist.get_playlists("All Tracks")
    assert all_tracks_playlist
    all_tracks_playlist = all_tracks_playlist[0]

    # Since the "Tracks" folder contains one track for each track in the
    # collection, the aggregated "All Tracks" playlist will contain all of
    # these tracks.
    tracks_in_playlist = all_tracks_playlist.get_tracks()
    for track_id in tracks:
        assert track_id in tracks_in_playlist


@pytest.mark.parametrize(
    "playlist_content,expected_tags,expected_tracks",
    [
        # Test only numerical selectors with existing values are added.
        (
            "[0] | [2] | [140] | [6] | [2021] | [1000]",
            ["[0]", "[140]", "[2021]"],
            [{"2", "3"}, {"1"}, {"3"}],
        ),
        # Do the same for ranges as well.
        (
            "[0-5] | [80-180] | [80-81] | [2021-2023] | [1000-1001]",
            ["[0-5]", "[80-180]", "[2021-2023]"],
            [{"1", "2", "3", "4"}, {"1", "2", "3", "4"}, {"1", "2", "3", "4"}],
        ),
        # Test only string selectors with existing values are added.
        (
            (
                "{artist:Biome} | {artist:Not there} | "
                "{comment:Whatta classic} | {comment:Not there} | "
                "{date:2021} | {date:Not there} | "
                "{key:4A} | {key:Not there} | "
                "{label:Label} | {label:Not there}"
            ),
            [
                "{artist:Biome}",
                "{comment:Whatta classic}",
                "{date:2021}",
                "{key:4A}",
                "{label:Label}",
            ],
            [{"1"}, {"2"}, {"3"}, {"4"}, {"1", "2"}],
        ),
        # Test that globbing works for string selectors.
        (
            "{artist:*Ca*} | {comment:*a*} | {key:7*} | {key:*A} | {label:*label}",
            [
                "{artist:*Ca*}",
                "{comment:*a*}",
                "{key:7*}",
                "{key:*A}",
                "{label:*label}",
            ],
            [
                {"2", "3"},
                {"1", "2", "4"},
                {"1", "2"},
                {"1", "3", "4"},
                {"1", "2", "3"},
            ],
        ),
        # Test that date added selections work as expected.
        (
            "{date:2022} | {date:>2022} | {date:>=2022} | {date:<2022} | {date:<=2022}",
            [
                "{date:2022}",
                "{date:>2022}",
                "{date:>=2022}",
                "{date:<2022}",
                "{date:<=2022}",
            ],
            [{"2"}, {"1", "4"}, {"1", "2", "4"}, {"3"}, {"2", "3"}],
        ),
        # Test that playlist selections work.
        (
            "{playlist:Hip Hop} & [0]",
            ["{playlist:Hip Hop}", "[0]"],
            [{"2"}],
        ),
        # Test that timedelta evaluate properly.
        (
            "{date:>1m}",
            ["{date:>1m}"],
            [{"1", "4"}],
        ),
    ],
)
def test_add_selectors_to_tags(
    playlist_content, expected_tags, expected_tracks, rekordbox_collection
):
    """Test for the add_selectors_to_tags function."""
    tags_tracks = defaultdict(dict)
    with mock.patch("djtools.collection.helpers.datetime") as mock_datetime:
        mock_datetime.strptime.side_effect = datetime.strptime
        mock_datetime.now.return_value = datetime(2023, 6, 25, 0, 0)
        add_selectors_to_tags(
            playlist_content, tags_tracks, rekordbox_collection, []
        )
    assert set(expected_tags) == set(tags_tracks)
    for tag, tracks in zip(expected_tags, expected_tracks):
        assert set(tags_tracks[tag]) == tracks


@pytest.mark.parametrize(
    "playlist_selector,in_tags_tracks",
    [
        ("{playlist:Genres}", False),
        ("{playlist:Hip Hop}", True),
    ],
)
def test_add_selectors_to_tags_skips_playlists_that_are_folders(
    playlist_selector, in_tags_tracks, rekordbox_collection
):
    """Test for the add_selectors_to_tags function."""
    playlist_content = f"{playlist_selector} & [0]"
    tags_tracks = defaultdict(dict)
    add_selectors_to_tags(
        playlist_content, tags_tracks, rekordbox_collection, []
    )
    assert (playlist_selector in tags_tracks) == in_tags_tracks


def test_parse_numerical_selectors():
    """Test for the parse_numerical_selectors function."""
    matches = ["1", "2-4", "140", "141-143", "2021", "2021-2023"]
    numerical_lookup = {}
    values = parse_numerical_selectors(matches, numerical_lookup)
    assert values == {
        "1",
        "2",
        "3",
        "4",
        "140",
        "141",
        "142",
        "143",
        "2021",
        "2022",
        "2023",
    }
    for match in matches:
        key = match
        if "-" in match:
            value_range = list(map(int, match.split("-")))
            value_range[-1] += 1
            key = tuple(map(str, range(*value_range)))
        assert key in numerical_lookup
        assert numerical_lookup[key] == f"[{match}]"


@pytest.mark.parametrize(
    "matches,expected",
    [
        # Numerical selectors work as follows:
        #  - value 0 through 5 represent ratings
        #  - values 6 through 999 represent BPMs
        #  - values 1000 and above represent years
        # Since this numerical selector spans 5 through 6, it tries to
        # represent both ratings and BPMs simultaneously which is invalid.
        (["5-6"], "Bad numerical range selector: 5-6"),
        # Since this numerical selector is, in fact, non-numerical, it's
        # considered malformed.
        (["bad"], "Malformed numerical selector: bad"),
    ],
)
def test_parse_numerical_selectors_warns_bad(matches, expected, caplog):
    """Test for the parse_numerical_selectors function."""
    caplog.set_level("WARNING")
    parse_numerical_selectors(matches, {})
    assert caplog.records[0].message == expected


def test_parse_string_selectors():
    """Test for the parse_string_selectors function."""
    matches = [
        "artist:*Tribe*",
        "comment:*Dark*",
        "date:2022",
        "date:<2022",
        "key:7A",
        "label:Some Label",
    ]
    string_lookup = {}
    type_map = {
        "artist": "get_artists",
        "comment": "get_comments",
        "date": "get_date_added",
        "key": "get_key",
        "label": "get_label",
    }
    playlists = set()
    parse_string_selectors(matches, string_lookup, type_map, playlists)
    for match in matches:
        key = tuple(map(str.strip, match.split(":")))
        if key[0] == "date":
            key = list(key)
            if re.search(DATE_SELECTOR_REGEX, key[1]):
                inequality, date = filter(
                    None, re.split(DATE_SELECTOR_REGEX, key[1])
                )
                key[1] = tuple(
                    [
                        INEQUALITY_MAP[inequality],
                        datetime.strptime(date, "%Y"),
                        "%Y",
                    ]
                )
            else:
                key[1] = tuple([None, datetime.strptime(key[1], "%Y"), "%Y"])
            key = tuple(key)
        assert key in string_lookup
        assert string_lookup[key] == f"{{{match}}}"


@pytest.mark.parametrize(
    "matches,expected",
    [
        # Valid string selectors include:
        #  - artist
        #  - comment
        #  - date
        #  - key
        #  - label
        (["bad thing:stuff"], "bad thing is not a supported selector!"),
        # Valid date formats include:
        #  -  YYYY-MM-DD
        #  -  YYYY-MM
        #  -  YYYY
        (["date:12345"], "Date selector 12345 is invalid!"),
    ],
)
def test_parse_string_selectors_warns_bad(matches, expected, caplog):
    """Test for the parse_string_selectors function."""
    caplog.set_level("WARNING")
    parse_string_selectors(matches, {}, {"date": "get_date_added"}, set())
    assert caplog.records[0].message == expected


@pytest.mark.parametrize("time_str", ["1y", "6m", "3m2w", "7d"])
def test_parse_timedelta(time_str):
    """Test for the parse_timedelta function."""
    with mock.patch("djtools.collection.helpers.datetime") as mock_datetime:
        mock_datetime.now.return_value = datetime(2024, 6, 22, 0, 0)
        relative_time = parse_timedelta(time_str)
        assert relative_time < mock_datetime.now.return_value


def test_parse_expression(rekordbox_track):
    """Test for the parse_expression function."""
    track_dict = {"2": rekordbox_track}
    expression = "Hip Hop & [0]"
    tags_tracks = {"Hip Hop": track_dict, "[0]": track_dict}
    tracks = parse_expression(expression, tags_tracks)
    assert tracks == track_dict


@pytest.mark.parametrize(
    "operators,tags,expected",
    [
        (
            ["&", "|", "~"],
            ["Jungle", "Breaks", "Techno", "Tech House"],
            {11, 12},
        ),
        (["~"], ["*House", "Bass House"], {3, 5, 6, 7, 8}),
        (["&"], ["{All DnB}", "Dark"], {2}),
    ],
)
def test_booleannode_evaluate(operators, tags, expected):
    """Test for the BooleanNode class."""
    tracks = {
        "{All DnB}": [1, 2, 3],
        "Acid House": [7, 8],
        "Bass House": [9, 10],
        "Breaks": [3, 4],
        "Dark": [2, 11],
        "Jungle": [1, 3],
        "Tech House": [3, 5, 6],
        "Techno": [11, 12],
    }
    tracks = {k: {x: None for x in v} for k, v in tracks.items()}
    node = BooleanNode(tracks)
    for operator in operators:
        node.add_operator(operator)
    for tag in tags:
        node.add_operand(tag)
    result = node.evaluate().keys()
    assert result == expected


@pytest.mark.parametrize(
    "track_ids,operator,expected",
    [
        ([1, 1], "&", {1}),
        ([1, 2], "&", {}),
        ([1, 1], "|", {1}),
        ([1, 2], "|", {1, 2}),
        ([1, 1], "~", {}),
        ([1, 2], "~", {1}),
    ],
)
def test_booleannode_evaluates_two_sets_of_tracks_with_each_kind_of_operator(
    track_ids, operator, expected
):
    """Test for the BooleanNode class."""
    node = BooleanNode({})
    for track_id in track_ids:
        node.add_operand({track_id: None})
    node.add_operator(operator)
    result = node.evaluate().keys()
    assert (result if result else dict(result)) == expected


@pytest.mark.parametrize(
    "tag,expected",
    [
        ("Techno", {1: None}),
        ("*Techno", {1: None, 2: None}),
        ("Techno*", {1: None, 3: None}),
        ("*Techno*", {1: None, 2: None, 3: None}),
    ],
)
def test_booleannode_gets_tracks(tag, expected):
    """Test for the BooleanNode class."""
    tags_tracks = {
        "Techno": {1: None},
        "Hard Techno": {2: None},
        "Techno Stuff": {3: None},
    }
    node = BooleanNode(tags_tracks)
    assert (
        node._get_tracks(tag) == expected  # pylint: disable=protected-access
    )


def test_booleannode_raises_runtime_eror():
    """Test for the BooleanNode class.

    A valid expression must contain one more operand than there are operators.
    """
    node = BooleanNode({})
    node.add_operator("|")
    node.add_operand("tag")
    with pytest.raises(
        RuntimeError,
        match=(
            "Invalid boolean expression:\n"
            + re.escape("\toperands: ['tag']\n")
            + re.escape("\toperators: ['union']")
        ),
    ):
        node.evaluate()


@mock.patch("djtools.collection.helpers.print_data")
def test_print_playlists_tag_statistics(
    mock_print_data, rekordbox_xml, capsys
):
    """Test for the print_playlists_tag_statistics function."""
    collection = RekordboxCollection(path=rekordbox_xml)
    playlists = collection.get_playlists("Genres")[0]
    print_playlists_tag_statistics(playlists)
    cap = capsys.readouterr()
    call_count = 0
    for playlist in playlists:
        assert f"\n{playlist.get_name()} tag statistics:" in cap.out
        for track in playlist.get_tracks().values():
            genre_tags = track.get_genre_tags()
            other_tags = set(track.get_tags()).difference(set(genre_tags))
            if other_tags:
                assert "\nOther:\n" in cap.out
            if genre_tags:
                assert "\nGenre:\n" in cap.out
            call_count += int(bool(other_tags)) + int(bool(genre_tags))
    assert mock_print_data.call_count == call_count


@pytest.mark.parametrize(
    "maximum,expected",
    [
        (
            25,
            {
                "Aggro": 25,
                "Bounce": 2,
                "Chill": 1,
                "Dark": 16,
                "Melodic": 8,
                "Rave": 11,
            },
        ),
        (
            20,
            {
                "Aggro": 20,
                "Bounce": 1,
                "Chill": 1,
                "Dark": 13,
                "Melodic": 6,
                "Rave": 9,
            },
        ),
        (
            10,
            {
                "Aggro": 10,
                "Bounce": 1,
                "Chill": 1,
                "Dark": 6,
                "Melodic": 3,
                "Rave": 4,
            },
        ),
    ],
)
def test_scale_output(maximum, expected):
    """Test for the scale_output function."""
    data = {
        "Aggro": 30,
        "Bounce": 2,
        "Chill": 1,
        "Dark": 19,
        "Melodic": 9,
        "Rave": 13,
    }
    scaled_data = scale_data(data, maximum=maximum)
    assert scaled_data == expected


def test_print_data(capsys):
    """Test for the print_data function."""
    data = {
        "Aggro": 30,
        "Bounce": 2,
        "Chill": 0,
        "Dark": 19,
        "Melodic": 9,
        "Rave": 13,
    }
    expected = (
        "|   *                                     \n"
        + "|   *                                     \n"
        + "|   *                                     \n"
        + "|   *                                     \n"
        + "|   *                                     \n"
        + "|   *                                     \n"
        + "|   *                                     \n"
        + "|   *                                     \n"
        + "|   *                                     \n"
        + "|   *               *                     \n"
        + "|   *               *                     \n"
        + "|   *               *                     \n"
        + "|   *               *                     \n"
        + "|   *               *                     \n"
        + "|   *               *                 *   \n"
        + "|   *               *                 *   \n"
        + "|   *               *                 *   \n"
        + "|   *               *        *        *   \n"
        + "|   *               *        *        *   \n"
        + "|   *               *        *        *   \n"
        + "|   *               *        *        *   \n"
        + "|   *               *        *        *   \n"
        + "|   *               *        *        *   \n"
        + "|   *       *       *        *        *   \n"
        + "|   *       *       *        *        *   \n"
        + "------------------------------------------\n"
        + "  Aggro   Bounce   Dark   Melodic   Rave  \n"
    )
    print_data(data)
    cap = capsys.readouterr()
    assert cap.out == expected



"""This script is used to generate histograms of My Tags per user."""

# pylint: disable=import-error,redefined-outer-name,unused-argument,invalid-name
from argparse import ArgumentParser
from collections import defaultdict
from itertools import groupby
from pathlib import Path
from typing import Optional, Set

import matplotlib.pyplot as plt
from matplotlib.ticker import MaxNLocator

from djtools.configs import build_config
from djtools.collection.base_collection import Collection
from djtools.collection.platform_registry import PLATFORM_REGISTRY


def analyze_collection_vibes(
    collection: Collection,
    included_tags: Optional[Set] = None,
    **kwargs,
):
    """Create histograms that show a collection's distribution of My Tags.

    Args:
        collection: Collection object.
        included_tags: My Tags to include in the histograms.
    """
    tracks = collection.get_tracks().values()
    tag_counts = {tag: 0 for tag in included_tags}
    for track in tracks:
        tags = set(track.get_tags()).difference(track.get_genre_tags())
        for tag in tags:
            if tag not in tag_counts:
                continue
            tag_counts[tag] += 1

    fig, ax = plt.subplots()
    fig.suptitle("Collection Vibes", fontsize=16, fontweight="extra bold")
    fig.text(0.5, 0.04, "Vibes", fontsize=16, ha="center", fontweight="bold")
    fig.text(
        0.04,
        0.5,
        "Number of Tags",
        fontsize=16,
        va="center",
        rotation="vertical",
        fontweight="bold",
    )
    x_ticks = list(range(len(tag_counts)))
    ax.bar(x_ticks, list(tag_counts.values()))
    ax.set_xticks(x_ticks)
    ax.set_xticklabels(list(tag_counts.keys()), rotation=70, fontsize=14)
    plt.tight_layout()
    plt.show()


def analyze_user_vibes(
    collection: Collection,
    user_index: int,
    included_tags: Optional[Set] = None,
    verbosity: int = 0,
    **kwargs,
):
    """Create histograms that show each user's distribution of My Tags.

    Args:
        collection: Collection object.
        user_index: Part of the `Location` path the corresponds to a username.
        included_tags: My Tags to include in the histograms.
        verbosity: Verbosity level.
    """
    tracks = collection.get_tracks().values()

    # Collect user-tag data.
    included_tags = included_tags or set()
    user_tags = defaultdict(lambda: defaultdict(int))
    for user, tracks in groupby(
        sorted(tracks, key=lambda x: x.get_location()),
        key=lambda x: list(filter(None, str(x.get_location()).split("/")))[
            user_index
        ],
    ):
        tagged_tracks = 0
        tracks = list(tracks)
        for track in tracks:
            tags = (
                set(track.get_tags())
                .difference(track.get_genre_tags())
                .intersection(included_tags)
            )
            if not tags:
                continue
            tagged_tracks += 1
            for tag in tags:
                user_tags[user][tag] += 1
        if verbosity:
            print(
                f"{user}\n\tTracks: {len(tracks)}\n\tTagged tracks: "
                f"{tagged_tracks}\n\tTags: {sum(user_tags[user].values())}"
            )

    # Plot bar charts for each user.
    rows = len(user_tags) // 2
    fig, ax = plt.subplots(rows, 2)  # pylint: disable=invalid-name
    fig.suptitle("User Vibes", fontsize=16, fontweight="extra bold")
    fig.text(0.5, 0.04, "Vibes", fontsize=16, ha="center", fontweight="bold")
    fig.text(
        0.04,
        0.5,
        "Number of Tags",
        fontsize=16,
        va="center",
        rotation="vertical",
        fontweight="bold",
    )
    # included_tags = sorted(
    #     set(key for tag in user_tags.values() for key in tag.keys())
    # )
    x_ticks = list(range(len(included_tags)))
    for idx, (user, tags) in enumerate(user_tags.items()):
        tags.update({vibe: 0 for vibe in included_tags if vibe not in tags})
        ax[idx % 2, int(idx < rows)].bar(
            x_ticks, [tags[vibe] for vibe in included_tags]
        )
        ax[idx % 2, int(idx < rows)].set_title(f"{user}", fontsize=14)
        ax[idx % 2, int(idx < rows)].set_xticks(x_ticks)
        ax[idx % 2, int(idx < rows)].yaxis.set_major_locator(
            MaxNLocator(integer=True)
        )
        if idx % 2 == rows - 1:
            ax[idx % 2, int(idx < rows)].set_xticklabels(
                included_tags, rotation=70, fontsize=14
            )
        else:
            ax[idx % 2, int(idx < rows)].set_xticklabels([])
    plt.tight_layout()
    plt.show()


if __name__ == "__main__":
    parser = ArgumentParser()
    parser.add_argument(
        "--collection",
        type=str,
        required=False,
        help="Path to collection.",
    )
    parser.add_argument(
        "--config",
        type=str,
        required=True,
        help="Path to config.yaml.",
    )
    parser.add_argument(
        "--included-tags",
        nargs="+",
        default=[
            "Aggro",
            "Heavy",
            "Rave",
            "Gangsta",
            "Bounce",
            "Groovy",
            "Deep",
            "Atmospheric",
            "Hypnotic",
            "Dark",
            "Melancholy",
            "Melodic",
            "Uplifting",
            "Strange",
        ],
        help="Included tags",
    )
    parser.add_argument(
        "--mode",
        choices=["collection", "user"],
        default="collection",
        help="Whether to analyze vibes across the whole collection or on a per-user level.",
    )
    parser.add_argument(
        "--user-index",
        type=int,
        default=3,
        help="Index of path part corresponding to username",
    )
    parser.add_argument(  # pylint: disable=duplicate-code
        "--verbosity",
        "-v",
        action="count",
        default=0,
        help="verbosity",
    )
    args = parser.parse_args()

    # Load config and, if provided, override path to collection.
    config_path = Path(args.config)
    config = build_config(config_path)

    # Load collection and get a dict of tracks keyed by location.
    collection = PLATFORM_REGISTRY[config.platform]["collection"](
        path=args.collection or config.collection_path
    )
    args.collection = collection

    if args.mode == "collection":
        analyze_collection_vibes(**vars(args))
    elif args.mode == "user":
        analyze_user_vibes(**vars(args))



"""Cluster tracks by tags."""

# pylint: disable=import-error,redefined-outer-name,duplicate-code
from argparse import ArgumentParser
from pathlib import Path
from typing import Dict, List, Tuple, Optional

import numpy as np
import pandas as pd
from sklearn.cluster import DBSCAN, KMeans, SpectralClustering
from tqdm import tqdm

from djtools.collection.base_collection import Collection
from djtools.collection.base_playlist import Playlist
from djtools.collection.base_track import Track
from djtools.collection.platform_registry import PLATFORM_REGISTRY
from djtools.configs import build_config


EXCLUDE_TAGS = (
    "DELETE",
    "Flute",
    "Guitar",
    "Horn",
    "Piano",
    "Scratch",
    "Strings",
    "Vocal",
)
PLAYLIST_NAME = "Clusters"


def get_collection_and_playlist_class(
    config_path: Path, collection_path: Optional[Path] = None
) -> Tuple[Collection, Playlist]:
    """Get the Collection and Playlist class for the configured platform.

    Args:
        config_path: Path to config.yaml.
        collection_path: Path to collection.

    Returns:
        Tuple of Collection object and Playlist class.
    """
    config = build_config(config_path)
    collection = PLATFORM_REGISTRY[config.platform]["collection"](
        path=collection_path or config.collection_path
    )
    playlist_class = PLATFORM_REGISTRY[config.platform]["playlist"]

    return collection, playlist_class


def get_tracks(
    collection: Collection, playlists: Optional[List[str]] = None
) -> Dict[str, Track]:
    """Get tracks for the provided list of playlists.

    Args:
        collection: Collection object.
        playlists: List of playlist names.

    Returns:
        Dictionary of tracks.
    """
    if not playlists:
        tracks = collection.get_tracks()
    else:
        tracks = {}
        for playlist_name in playlists:
            for playlist in collection.get_playlists(playlist_name):
                tracks.update(playlist.get_tracks())

    return tracks


def dataprep(collection: Collection, tracks: Dict[str, Track]) -> pd.DataFrame:
    """Build a Dataframe from the track tags.

    Args:
        collection: Collection object.
        tracks: Dictionary of tracks.

    Returns:
        Dataframe of track tag data.
    """
    other_tags = sorted(
        set(collection.get_all_tags()["other"]).difference(EXCLUDE_TAGS)
    )
    dataset = []
    for track in tqdm(tracks.values(), desc="Building dataset from tracks"):
        track_id = track.get_id()
        tags = set(track.get_tags()).difference(set(track.get_genre_tags()))
        one_hot = [0] * len(other_tags)
        for index, tag in enumerate(other_tags):
            if tag not in tags:
                continue
            one_hot[index] = 1
        dataset.append((track_id, *one_hot))
    data = pd.DataFrame(dataset, columns=["id"] + other_tags, index=None)
    data.set_index("id", inplace=True)

    return data


def cluster(
    data: pd.DataFrame,
    clusters: List[int],
    playlist_class: Playlist,
    tracks: Dict[str, Track],
    cluster_algo_name: str,
) -> List[Playlist]:
    """Create playlists from tracks clustered by tag data.

    Args:
        data: Dataframe of track tag data.
        clusters: Number of clusters to build.
        playlist_class: Playlist class.
        tracks: Dictionary of tracks.
        cluster_algo_name: Clustring algorithm name.

    Returns:
        List of Playlist objects.
    """
    playlists = []
    cluster_algos = {
        "kmeans": KMeans,
        "dbscan": DBSCAN,
        "spectral": SpectralClustering,
    }

    for num_clusters in clusters:
        if cluster_algo_name in ["kmeans", "spectral"]:
            kwargs = {"n_clusters": num_clusters, "n_init": 10}
        else:
            kwargs = {"eps": num_clusters, "min_samples": 10}
        cluster_playlists = []
        cluster_algo = cluster_algos[cluster_algo_name](**kwargs).fit(data)
        if getattr(cluster_algo, "inertia_", None):
            print(
                f"{num_clusters} clusters -- inertia: {cluster_algo.inertia_}:"
            )
        else:
            print(f"{num_clusters} clusters")
        for cluster in range(num_clusters):
            indices = np.where(cluster_algo.labels_ == cluster)
            track_ids = data.iloc[indices].index.tolist()
            print(f"\t{len(track_ids)}")
            cluster_playlists.append(
                playlist_class.new_playlist(
                    name=str(cluster + 1),
                    tracks={
                        track_id: tracks[track_id] for track_id in track_ids
                    },
                )
            )
        playlists.append(
            playlist_class.new_playlist(
                name=f"{num_clusters} clusters", playlists=cluster_playlists
            )
        )

    return playlists


def save_playlists(
    collection: Collection, playlists: List[Playlist], playlist_class: Playlist
):
    """Save Playlists to Collection.

    Args:
        collection: Collection object.
        playlists: List of playlist names.
        playlist_class: Playlist class.
    """
    previous_playlists = collection.get_playlists(name=PLAYLIST_NAME)
    root = collection.get_playlists()
    for playlist in previous_playlists:
        root.remove_playlist(playlist)
    collection.add_playlist(
        playlist_class.new_playlist(name=PLAYLIST_NAME, playlists=playlists)
    )
    collection.serialize()


if __name__ == "__main__":
    parser = ArgumentParser()
    parser.add_argument(
        "--cluster_algo_name",
        type=str,
        default="kmeans",
        help="Clustering algorithm to use.",
    )
    parser.add_argument(
        "--clusters",
        type=int,
        nargs="+",
        default=[5],
        help="Number of clusters.",
    )
    parser.add_argument(
        "--collection", required=False, help="Path to collection."
    )
    parser.add_argument("--config", required=True, help="Path to config.yaml.")
    parser.add_argument(
        "--playlists", nargs="+", help="Playlist names to cluster."
    )
    args = parser.parse_args()

    collection, playlist_class = get_collection_and_playlist_class(
        Path(args.config), Path(args.collection) if args.collection else None
    )
    tracks = get_tracks(collection, args.playlists)
    dataset = dataprep(collection, tracks)
    playlists = cluster(
        dataset, args.clusters, playlist_class, tracks, args.cluster_algo_name
    )
    save_playlists(collection, playlists, playlist_class)



"""Testing for the tracks module."""

import os
from datetime import datetime
from pathlib import Path

import pytest

from djtools.collection.rekordbox_track import RekordboxTrack


@pytest.mark.parametrize(
    "method,expected",
    [
        ("get_artists", "A Tribe Called Quest"),
        ("get_bpm", 86),
        ("get_comments", " /* Gangsta */ "),
        ("get_date_added", datetime(2022, 6, 24)),
        ("get_genre_tags", ["Hip Hop", "R&B"]),
        ("get_id", "2"),
        ("get_key", "7B"),
        ("get_label", "Label"),
        ("get_location", "track2.mp3"),
        ("get_rating", 0),
        ("get_tags", ["Hip Hop", "R&B", "Gangsta"]),
        ("get_year", "2022"),
    ],
)
def test_rekordboxtrack_get_methods(method, expected, rekordbox_track):
    """Test RekordboxTrack class."""
    try:
        _method = getattr(rekordbox_track, method)
    except AttributeError:
        assert False, f"RekordboxTrack is missing required method '{method}'"

    if method == "get_location":
        assert _method().name == expected
    else:
        assert _method() == expected


def test_rekordboxtrack_serialization(rekordbox_track_tag):
    """Test RekordboxTrack class."""
    track = RekordboxTrack(rekordbox_track_tag)
    loc = str(track.get_location())
    assert (
        repr(track)
        == """RekordboxTrack(\n    Artist="A Tribe Called Quest", """
        """AverageBpm=86.0, Comments=" /* Gangsta */ ", """
        """\n    DateAdded="2022-06-24", Genre=['Hip Hop', 'R&B'], """
        f"""Label="Label", Location={loc}, \n    Tonality="7B", Rating=0, """
        """TrackID="2", TrackNumber=2, Year="2022", MyTags=['Gangsta'], """
        """\n    Tags=['Hip Hop', 'R&B', 'Gangsta'], beat_grid=0, hot_cues=0\n)"""
    )
    sep = "" if os.name == "posix" else "/"  # pylint: disable=no-member
    loc = track.get_location().as_posix()
    assert (
        str(track)
        == """<TRACK Artist="A Tribe Called Quest" AverageBpm="86.00" """
        """Comments=" /* Gangsta */ " DateAdded="2022-06-24" """
        """Genre="Hip Hop / R&amp;B" Label="Label" """
        f"""Location="file://localhost{sep}{loc}" Rating="0" Tonality="7B" """
        """TrackID="2" TrackNumber="2" Year="2022"/>"""
    )
    assert track.serialize() == rekordbox_track_tag


def test_rekordboxtrack_set_location(rekordbox_track_tag):
    """Test RekordboxTrack class."""
    track = RekordboxTrack(rekordbox_track_tag)
    track.set_location("path/to.mp3")
    assert track.get_location() == Path("path/to.mp3")


def test_rekordboxtrack_set_track_number(rekordbox_track_tag):
    """Test RekordboxTrack class."""
    track = RekordboxTrack(rekordbox_track_tag)
    track_number = 42
    track.set_track_number(track_number)
    assert f'TrackNumber="{track_number}"' in str(track)



"""Script for analyzing timing of unit tests and fixtures."""

# pylint: disable=import-error
from collections import defaultdict
from itertools import groupby

import numpy as np
import pyperclip


TIME_LOWER_BOUND = 0.1


data = pyperclip.paste()
test_case_lines = []
fixture_lines = []
for line in data.split("\n"):
    if "[DEBUG]" not in line:
        continue
    if "AssertionError" in line:
        continue
    if "[fixture]" in line:
        fixture_lines.append(line)
    else:
        test_case_lines.append(line)

test_lookup = {}


def func(string):
    """Helper function for sorting and grouping test results."""
    return string.split("[DEBUG] ")[-1].split(" [")[0].split("[")[0].strip()


for _id, group in groupby(sorted(test_case_lines, key=func), key=func):
    phase_times = defaultdict(list)
    for item in group:
        item = item.split(" [")[-1]
        phase, time = item.split("]=")
        phase_times[phase].append(float(time))
    for key, value in phase_times.items():
        phase_times[key] = np.mean(value)
    test_lookup[_id] = dict(phase_times)


sorted_test_lookup = {
    phase: {
        key: value[phase]
        for key, value in sorted(
            test_lookup.items(),
            key=lambda x: x[1][phase],  # pylint: disable=cell-var-from-loop
            reverse=True,
        )
    }
    for phase in ["call", "setup", "teardown"]
}

for phase, times in sorted_test_lookup.items():
    print(phase)
    for test, time in times.items():
        if time < TIME_LOWER_BOUND:
            continue
        print(f"{test}: {time}")
    print()


fixture_times = {}


def func(string):  # pylint: disable=function-redefined
    """Helper function for sorting and grouping test results."""
    return string.split("[DEBUG] ")[-1].split(" [fixture]=")[0]


for _id, group in groupby(sorted(fixture_lines, key=func), key=func):
    group = [float(x.strip().split("[fixture]=")[-1]) for x in group]
    fixture_times[_id] = np.mean(group)


for fixture, time in dict(
    sorted(fixture_times.items(), key=lambda x: x[1], reverse=True)
).items():
    print(f"{fixture}: {time}")
print()



"""Testing for the playlist_builder module."""

from unittest import mock

import pytest

from djtools.collection.config import (
    PlaylistConfig,
    PlaylistConfigContent,
    PlaylistRemainder,
)
from djtools.collection.playlist_builder import (
    collection_playlists,
    PLAYLIST_NAME,
)
from djtools.collection.rekordbox_collection import RekordboxCollection
from djtools.collection.rekordbox_playlist import RekordboxPlaylist


@pytest.mark.parametrize(
    "remainder_type",
    [PlaylistRemainder.FOLDER, PlaylistRemainder.PLAYLIST],
)
def test_collection_playlists_makes_unused_tags_playlists(
    remainder_type,
    config,
    rekordbox_collection,
    rekordbox_xml,
    playlist_config_obj,
):
    """Test for the collection_playlists function."""
    config.collection.collection_path = rekordbox_xml
    config.collection.collection_playlists_remainder = remainder_type
    new_path = rekordbox_xml.parent / "test_collection"

    tags = {
        tag
        for value in rekordbox_collection.get_all_tags().values()
        for tag in value
    }
    some_tag = next(iter(tags))
    tags.remove(some_tag)

    playlist_config_obj.combiner = None
    playlist_config_obj.tags.playlists = [some_tag]
    config.collection.playlist_config = playlist_config_obj

    collection_playlists(config, path=new_path)

    # Since our "tags" config only specifies one tag, and there is more than
    # one tag in the collection, there must be a playlist called "Unused Tags".
    collection = RekordboxCollection(new_path)
    assert collection.get_playlists("Unused Tags")
    unused_tags_playlists = collection.get_playlists("Unused Tags")[0]

    if remainder_type == PlaylistRemainder.FOLDER:
        # If collection_playlists_remainder is set to "folder", then
        # "Unused Tags" will be a folder containing one playlist for each
        # unused tag.
        assert unused_tags_playlists.is_folder()

        for playlist in unused_tags_playlists:
            tag_playlist = playlist.get_name()

            # The name of the playlist will be one of the tags in the
            # collection.
            assert tag_playlist in tags

            # The name of the playlist will be a tag that's not specified in
            # the "tags" section of the playlist config.
            assert tag_playlist != some_tag

            # Every track in that playlist will have the tag that is the name
            # of that playlist.
            assert all(
                tag_playlist in track.get_tags()
                for track in playlist.get_tracks().values()
            )
    else:
        # If collection_playlists_remainder is set to "playlist", then
        # "Unused Tags" will be a playlist containing all the tracks having at
        # least one of the unused tags.
        assert not unused_tags_playlists.is_folder()

        for track in unused_tags_playlists.get_tracks().values():
            track_tags = track.get_tags()
            for tag in track_tags:
                if tag == some_tag:
                    # If the tag that was specified in the "tags" section of
                    # the playlist config appears in this track, then it must
                    # be the case that it also has other tags which are unused.
                    assert len(track_tags) > 1
                    continue

                # Otherwise, the tag must be one of the unused tags.
                assert tag in tags


@mock.patch(
    "djtools.collection.playlist_builder.print_playlists_tag_statistics"
)
def test_collection_playlists_prints_playlist_tag_statistics(
    mock_print_playlists_tag_statistics,
    config,
    rekordbox_xml,
    playlist_config_obj,
):
    """Test for the collection_playlists function."""
    config.collection.collection_path = rekordbox_xml
    config.verbosity = 1
    config.collection.playlist_config = playlist_config_obj

    collection_playlists(config, path=rekordbox_xml.parent / "test_collection")

    assert mock_print_playlists_tag_statistics.call_count == 1


@pytest.mark.parametrize(
    "invalid_expression",
    [
        "this & & will be invalid",
        {
            "name": "invalid expression",
            "tag_content": "this & & will be invalid",
        },
    ],
)
def test_collection_playlists_handles_error_parsing_expression(
    invalid_expression, config, rekordbox_xml, playlist_config_obj, caplog
):
    """Test for the collection_playlists function."""
    caplog.set_level("WARNING")
    config.collection.collection_path = rekordbox_xml
    playlist_config_obj.tags = None
    playlist_config_obj.combiner.playlists = [
        PlaylistConfigContent(
            name="test",
            playlists=[invalid_expression],
        )
    ]
    config.collection.playlist_config = playlist_config_obj

    collection_playlists(config, path=rekordbox_xml.parent / "test_collection")

    assert caplog.records[0].message.startswith(
        "Error parsing expression: this & & will be invalid"
    )


def test_collection_playlists_removes_existing_playlist(
    config, playlist_config_obj, rekordbox_xml
):
    """Test for the collection_playlists function."""
    collection = RekordboxCollection(rekordbox_xml)

    # A playlist_builder playlist should not already exist.
    assert not collection.get_playlists(PLAYLIST_NAME)

    # Insert a playlist_builder playlist into the collection.
    new_playlists = RekordboxPlaylist.new_playlist(PLAYLIST_NAME, playlists=[])
    root = collection.get_playlists()
    root.add_playlist(new_playlists)

    # A playlist_builder playlist should now exist.
    assert collection.get_playlists(PLAYLIST_NAME)

    # Serialize the collection containing a playlist_builder playlist.
    new_path = rekordbox_xml.parent / "test_collection"
    collection.serialize(path=new_path)
    config.collection.collection_path = new_path
    config.collection.playlist_config = playlist_config_obj

    # Run the playlist_builder on this collection to test removing the existing
    # playlist_builder playlist.
    with mock.patch(
        "djtools.collection.rekordbox_collection.RekordboxCollection.add_playlist"
    ):
        collection_playlists(config, path=new_path)

    collection = RekordboxCollection(new_path)
    assert not collection.get_playlists(PLAYLIST_NAME)


def test_collection_playlists_with_empty_playlistconfig_returns_early(
    config, rekordbox_xml, caplog
):
    """Test for the collection_playlists function."""
    caplog.set_level("WARNING")
    config.collection.collection_path = rekordbox_xml
    config.collection.playlist_config = PlaylistConfig()
    collection_playlists(config, path=rekordbox_xml.parent / "test_collection")
    assert caplog.records[0].message == (
        "Not building playlists because the playlist config is empty."
    )



"""Testing for the process_recording module."""

from pathlib import Path
from unittest import mock

import pytest

from djtools.utils.normalize_audio import normalize


def test_normalize_handles_no_local_tracks(config):
    """Test for the normalize function."""
    config.utils.local_dirs = []
    with pytest.raises(
        RuntimeError,
        match=(
            "There are no local tracks; make sure local_dirs has one or "
            "more directories containing one or more tracks"
        ),
    ):
        normalize(config)


def test_normalize_handles_decode_error(config, tmpdir, caplog):
    """Test for the normalize function."""
    caplog.set_level("ERROR")
    tmpdir = Path(tmpdir)
    filename = tmpdir / "bad_audio_file.txt"
    filename.write_text("something")
    config.utils.local_dirs = [tmpdir]
    normalize(config)
    assert caplog.records[0].message.startswith(f"Couldn't decode {filename}:")


@pytest.mark.parametrize("target_headroom", [0, 2, 5])
@mock.patch("djtools.utils.normalize_audio.effects.normalize")
@mock.patch("djtools.utils.process_recording.AudioSegment.export", mock.Mock())
def test_normalize(
    mock_normalize, target_headroom, audio_file, config, input_tmpdir
):
    """Test for the normalize function."""
    config.utils.audio_headroom = target_headroom
    audio, _ = audio_file
    file_path = Path(input_tmpdir) / "file.wav"
    with (
        mock.patch(
            "djtools.utils.normalize_audio.AudioSegment.from_file"
        ) as mock_audio_segment,
        mock.patch(
            "djtools.utils.normalize_audio.get_local_tracks",
            mock.Mock(return_value={"playlist": [file_path]}),
        ),
    ):
        mock_audio_segment.return_value = audio
        normalize(config)
        if abs(audio.max_dBFS + target_headroom) > 0.001:
            mock_normalize.assert_called_once()
    # TODO(a-rich): Getting unexpected headroom readings after exporting.
    # See this issue:
    # https://stackoverflow.com/questions/76791317/pydub-how-to-retain-headroom-across-export-and-from-file"  pylint: disable=line-too-long
    # audio = AudioSegment.from_file(_file)
    # assert abs(audio.max_dBFS + target_headroom) < 0.001


@mock.patch(
    "djtools.utils.normalize_audio.utils.mediainfo",
    mock.Mock(side_effect=FileNotFoundError()),
)
@mock.patch("djtools.utils.normalize_audio.effects.normalize", mock.Mock())
@mock.patch("djtools.utils.process_recording.AudioSegment.export", mock.Mock())
def test_normalize_handles_missing_ffmpeg(
    config, audio_file, input_tmpdir, caplog
):
    """Test for the normalize function."""
    caplog.set_level("WARNING")
    config.utils.audio_headroom = 0.0
    audio, _ = audio_file
    with (
        mock.patch(
            "djtools.utils.normalize_audio.AudioSegment.from_file"
        ) as mock_audio_segment,
        mock.patch(
            "djtools.utils.normalize_audio.get_local_tracks",
            mock.Mock(
                return_value={"playlist": [Path(input_tmpdir) / "file.wav"]}
            ),
        ),
    ):
        mock_audio_segment.return_value = audio
        normalize(config)
        assert caplog.records[0].message.startswith(
            'Couldn\'t export file with ID3 tags; ensure "ffmpeg" is installed'
        )



"""Testing for the process_recording module."""

from pathlib import Path
from unittest import mock

import pytest
from pydub import AudioSegment

from djtools.utils.process_recording import process


@mock.patch(
    "djtools.utils.process_recording.get_spotify_tracks",
    mock.Mock(return_value={}),
)
@mock.patch("djtools.utils.helpers.AudioSegment.export", mock.Mock())
def test_process_handles_missing_or_empty_playlist(config):
    """Test for the process function."""
    config.utils.recording_playlist = "playlist"
    with pytest.raises(
        RuntimeError,
        match=(
            "There are no Spotify tracks; make sure "
            "download_spotify_playlist is a key from "
            "spotify_playlists.yaml"
        ),
    ):
        process(config)


@mock.patch("djtools.utils.helpers.os.utime", mock.Mock())
@mock.patch("djtools.utils.helpers.AudioSegment.export", mock.Mock())
@mock.patch(
    "djtools.utils.process_recording.get_spotify_tracks",
    mock.Mock(
        return_value={
            "playlist": [
                {
                    "track": {
                        "album": {
                            "name": "",
                            "release_date_precision": "year",
                            "release_date": "2023",
                            "label": "",
                        },
                        "artists": [{"name": "some artist"}],
                        "duration_ms": 10000,
                        "name": "some name",
                    },
                },
            ]
        }
    ),
)
@mock.patch(
    "djtools.utils.process_recording.AudioSegment.from_file",
    mock.Mock(return_value=AudioSegment.silent(duration=2000)),
)
@mock.patch("djtools.utils.process_recording.process_parallel", mock.Mock())
@mock.patch("djtools.utils.process_recording.trim_initial_silence")
@pytest.mark.parametrize("trim,expected", [(0, 0), (1, 1), ("auto", 1)])
def test_process_skips_trimming_initial_silence(
    mock_trim_initial_silence, trim, expected, config, tmpdir
):
    """Test for the process function."""
    config.utils.audio_destination = Path(tmpdir)
    config.utils.recording_file = "file.wav"
    config.utils.recording_playlist = "playlist"
    config.utils.trim_initial_silence = trim
    process(config)
    assert mock_trim_initial_silence.call_count == expected


@mock.patch("djtools.utils.helpers.AudioSegment.export", mock.Mock())
@mock.patch("djtools.utils.process_recording.get_spotify_tracks")
@mock.patch("djtools.utils.process_recording.AudioSegment.from_file")
@mock.patch("djtools.utils.helpers.os.utime", mock.Mock())
def test_process_warns_when_recording_is_too_short(
    mock_audio_segment, mock_spotify_tracks, config, caplog, tmpdir
):
    """Test for the process function."""
    config.utils.recording_file = "file.wav"
    config.utils.recording_playlist = "playlist"
    config.utils.audio_destination = Path(tmpdir)
    caplog.set_level("WARNING")
    mock_audio_segment.return_value = AudioSegment.silent(duration=90_000)
    mock_spotify_tracks.return_value = {
        "playlist": [
            {
                "track": {
                    "album": {
                        "name": "",
                        "release_date_precision": "year",
                        "release_date": "2023",
                        "label": "",
                    },
                    "artists": [{"name": "artist"}],
                    "duration_ms": 90_001,
                    "name": "some - bad name",
                },
            },
        ],
    }
    audio_duration = len(mock_audio_segment.return_value)
    playlist_duration = sum(
        (
            track["track"]["duration_ms"] + 500
            for playlist in mock_spotify_tracks.return_value.values()
            for track in playlist
        )
    )
    process(config)
    assert caplog.records[0].message == (
        f"file.wav has a duration of {audio_duration} milliseconds which is "
        "less than the sum of track lengths in the Spotify playlist playlist "
        f"which is {playlist_duration} milliseconds. Please confirm your "
        "recording went as expected!"
    )


@mock.patch("djtools.utils.helpers.AudioSegment.export", mock.Mock())
@mock.patch("djtools.utils.helpers.effects.normalize", mock.Mock())
@mock.patch(
    "djtools.utils.process_recording.get_spotify_tracks",
    mock.Mock(
        return_value={
            "playlist": [
                {
                    "track": {
                        "album": {
                            "name": "",
                            "release_date_precision": "year",
                            "release_date": "2023",
                            "label": "",
                        },
                        "artists": [{"name": "artist"}],
                        "duration_ms": 1000,
                        "name": "some - bad name",
                    },
                }
            ],
        },
    ),
)
@mock.patch(
    "djtools.utils.process_recording.AudioSegment.from_file",
    mock.Mock(return_value=AudioSegment.silent(duration=2000)),
)
@mock.patch("djtools.utils.helpers.os.utime", mock.Mock())
def test_process_warns_when_filename_is_malformed(config, tmpdir, caplog):
    """Test for the process function."""
    caplog.set_level("WARNING")
    config.utils.recording_playlist = "playlist"
    config.utils.audio_destination = Path(tmpdir)
    process(config)
    filename = tmpdir / "some - bad name - artist.mp3"
    assert caplog.records[0].message == (
        f'{filename} has more than one occurrence of " - "! '
        "Because djtools splits on this sequence of characters to "
        "separate track title and artist(s), you might get unexpected "
        'behavior while using features like "--check-tracks".'
    )


@mock.patch("djtools.utils.helpers.AudioSegment.export")
@mock.patch("djtools.utils.helpers.effects.normalize")
@mock.patch("djtools.utils.process_recording.AudioSegment.from_file")
@mock.patch("djtools.utils.process_recording.get_spotify_tracks")
def test_process(
    mock_get_spotify_tracks,
    mock_audio,
    mock_normalize,
    mock_export,
    config,
    tmpdir,
):
    """Test for the process function."""
    mock_get_spotify_tracks.return_value = {
        "playlist": [
            {
                "track": {
                    "album": {
                        "name": "",
                        "release_date_precision": "year",
                        "release_date": "2023",
                        "label": "",
                    },
                    "artists": [{"name": "some artist"}],
                    "duration_ms": 10000,
                    "name": "some name",
                },
            },
            {
                "track": {
                    "album": {
                        "name": "",
                        "release_date_precision": "month",
                        "release_date": "2023-07",
                        "label": "",
                    },
                    "artists": [
                        {"name": "another artist"},
                        {"name": "a final artist"},
                    ],
                    "duration_ms": 10000,
                    "name": "another name",
                },
            },
            {
                "track": {
                    "album": {
                        "name": "",
                        "release_date_precision": "day",
                        "release_date": "2023-07-28",
                        "label": "",
                    },
                    "artists": [
                        {"name": "another artist"},
                        {"name": "a final artist"},
                    ],
                    "duration_ms": 10000,
                    "name": "a final name",
                },
            },
        ],
    }
    mock_audio.return_value = AudioSegment.silent(duration=30000)
    mock_normalize.return_value = mock_audio.return_value

    def mock_export_function(
        filename, **kwargs
    ):  # pylint: disable=unused-argument
        with open(filename, mode="wb") as _file:
            _file.write(b"")

    mock_export.side_effect = mock_export_function
    config.utils.audio_destination = Path(tmpdir)
    config.utils.recording_file = "file.wav"
    config.utils.recording_playlist = "playlist"
    process(config)
    track_count = len(mock_get_spotify_tracks.return_value["playlist"])
    assert mock_export.call_count == track_count
    assert mock_normalize.call_count == track_count
    assert len(list(config.utils.audio_destination.iterdir())) == track_count



"""This module is used to compare tracks from Spotify playlists and / or local
directories to see if there is any overlap with the contents of the Beatcloud.
"""

import logging
from collections import defaultdict
from itertools import groupby
from operator import itemgetter
from typing import List, Optional, Tuple, Type

from djtools.utils.helpers import (
    find_matches,
    get_spotify_tracks,
    get_beatcloud_tracks,
    get_local_tracks,
    reverse_title_and_artist,
)


logger = logging.getLogger(__name__)
BaseConfig = Type["BaseConfig"]


def compare_tracks(
    config: BaseConfig,
    beatcloud_tracks: Optional[List[str]] = None,
) -> Tuple[List[str], List[str]]:
    """Compares tracks from Spotify / local with Beatcloud tracks.

    Gets track titles and artists from Spotify playlist(s) and / or file names
    from local directories, and get file names from the beatcloud. Then compute
    the Levenshtein similarity between their product in order to identify any
    overlapping tracks.

    Args:
        config: Configuration object.
        beatcloud_tracks: Cached list of tracks from S3.

    Returns:
        Tuple with a list of all Beatcloud tracks and list of full paths to
            matching Beatcloud tracks.
    """
    if config.sync.download_spotify_playlist:
        cached_local_dirs = config.utils.local_dirs
        config.utils.local_dirs = []
        spotify_playlists = [config.sync.download_spotify_playlist]
    else:
        spotify_playlists = config.utils.check_tracks_spotify_playlists

    track_sets = []
    beatcloud_matches = []
    if spotify_playlists:
        spotify_tracks = get_spotify_tracks(config, spotify_playlists)
        if not spotify_tracks:
            if config.sync.download_spotify_playlist:
                substring = "download_spotify_playlist is a key"
            else:
                substring = (
                    "check_tracks_spotify_playlists has one or more keys"
                )
            logger.warning(
                f"There are no Spotify tracks; make sure {substring} in "
                "spotify_playlists.yaml"
            )
        else:
            track_results = defaultdict(list)
            for playlist_name, playlist_tracks in spotify_tracks.items():
                for track in playlist_tracks:
                    title = track["track"]["name"]
                    artists = ", ".join(
                        [y["name"] for y in track["track"]["artists"]]
                    )
                    track_results[playlist_name].append(
                        f"{artists} - {title}"
                        if config.sync.artist_first
                        else f"{title} - {artists}"
                    )
            track_sets.append((track_results, "Spotify Playlist Tracks"))
    if config.utils.local_dirs:
        local_tracks = get_local_tracks(config)
        if not local_tracks:
            logger.warning(
                "There are no local tracks; make sure local_dirs has one or "
                "more directories containing one or more tracks"
            )
        else:
            track_results = {
                key: [track.stem for track in value]
                for key, value in local_tracks.items()
            }
            track_sets.append((track_results, "Local Directory Tracks"))

    if config.sync.download_spotify_playlist:
        config.utils.local_dirs = cached_local_dirs

    if not track_sets:
        return beatcloud_tracks, beatcloud_matches

    if not beatcloud_tracks:
        beatcloud_tracks = get_beatcloud_tracks(config.sync.bucket_url)

    path_lookup = {x.stem: x for x in beatcloud_tracks}

    for tracks, track_type in track_sets:
        if config.sync.artist_first and track_type == "Local Directory Tracks":
            path_lookup = reverse_title_and_artist(path_lookup)
        matches = find_matches(
            tracks,
            path_lookup.keys(),
            config,
        )
        logger.info(f"\n{track_type} / Beatcloud Matches: {len(matches)}")
        for loc, matches in groupby(
            sorted(matches, key=itemgetter(0)), key=itemgetter(0)
        ):
            logger.info(f"{loc}:")
            for _, track, beatcloud_track, fuzz_ratio in matches:
                beatcloud_matches.append(path_lookup[beatcloud_track])
                logger.info(f"\t{fuzz_ratio}: {track} | {beatcloud_track}")

    return beatcloud_tracks, beatcloud_matches



"""Testing for the shuffle_playlists module."""

import pytest

from djtools.collection.base_playlist import Playlist
from djtools.collection.rekordbox_collection import RekordboxCollection
from djtools.collection.shuffle_playlists import shuffle_playlists


def test_shuffle_playlists_handles_missing_playlist(config, rekordbox_xml):
    """Test shuffle_playlists function."""
    playlist = "nonexistent playlist"
    config.collection.collection_path = rekordbox_xml
    config.collection.shuffle_playlists = [playlist]
    with pytest.raises(
        LookupError,
        match=f"{playlist} not found",
    ):
        shuffle_playlists(config)


def test_shuffle_playlists_shuffles_track_numbers(
    config, rekordbox_collection, rekordbox_xml, tmpdir
):
    """Test shuffle_playlists function."""
    playlist = "Hip Hop"
    config.collection.collection_path = rekordbox_xml
    config.collection.shuffle_playlists = [playlist]
    hip_hop_playlist = rekordbox_collection.get_playlists(playlist)[0]
    old_track_id_number_map = {
        track_id: track._TrackNumber  # pylint: disable=protected-access
        for track_id, track in hip_hop_playlist.get_tracks().items()
    }
    new_collection = tmpdir / "test_shuffle_collection"
    shuffle_playlists(config, path=new_collection)
    collection = RekordboxCollection(new_collection)
    hip_hop_playlist = collection.get_playlists(playlist)[0]
    new_track_id_number_map = {
        track_id: track._TrackNumber  # pylint: disable=protected-access
        for track_id, track in hip_hop_playlist.get_tracks().items()
    }
    assert old_track_id_number_map.keys() == new_track_id_number_map.keys()
    assert old_track_id_number_map.values() != new_track_id_number_map.values()


def test_shuffle_playlists_creates_new_playlist(
    config, rekordbox_collection, rekordbox_xml, tmpdir
):
    """Test shuffle_playlists function."""
    target_playlist = "Hip Hop"
    output_playlist = "SHUFFLE"
    config.collection.collection_path = rekordbox_xml
    config.collection.shuffle_playlists = [target_playlist]
    shuffle_playlist = rekordbox_collection.get_playlists(output_playlist)
    assert not shuffle_playlist
    new_collection = tmpdir / "test_collection"
    shuffle_playlists(config, path=new_collection)
    collection = RekordboxCollection(new_collection)
    shuffle_playlist = collection.get_playlists(output_playlist)[0]
    assert isinstance(shuffle_playlist, Playlist)


def test_shuffle_playlists_creates_new_collection(
    config, rekordbox_xml, tmpdir
):
    """Test shuffle_playlists function."""
    playlist = "Hip Hop"
    config.collection.collection_path = rekordbox_xml
    config.collection.shuffle_playlists = [playlist]
    new_collection = tmpdir / "test_collection"
    assert not new_collection.exists()
    shuffle_playlists(config, path=new_collection)
    assert new_collection.exists()



"""This module is a script for modifying a Collection.

More specifically, it queries Spotify for each track and attempts to resolve
the year, album, and label.
"""

# pylint: disable=redefined-outer-name,duplicate-code,protected-access,invalid-name
from argparse import ArgumentParser
from concurrent.futures import as_completed, ThreadPoolExecutor
from datetime import datetime
import os
from typing import Dict, Union

from tqdm import tqdm

from djtools.configs.helpers import build_config
from djtools.collection.base_track import Track
from djtools.collection.platform_registry import PLATFORM_REGISTRY
from djtools.spotify.helpers import filter_results, get_spotify_client


def get_spotify_tags_thread(track, spotify, threshold, query_limit):
    """Threaded function.

    Hits the Spotify API and tries to find a matching track. If found, resolve
    the year, album, and label fields of the track.

    Args:
        track: A Track object.
        spotify: Spotify client.
        threshold: Minimum Spotify result similarity for a match.
        query_limit: Number of Spotify query results.
    """
    title = getattr(track, "_Name")
    artist = track.get_artists()
    results = spotify.search(
        q=f"track:{title} artist:{artist}", type="track", limit=query_limit
    )
    result, _ = filter_results(spotify, results, threshold, title, artist)
    if result:
        album = spotify.album(result["album"]["id"])
        for date_format in ["%Y-%m-%d", "%Y-%m", "%Y"]:
            try:
                date = datetime.strptime(album["release_date"], date_format)
            except ValueError:
                continue
        for attribute_name, attribute in [
            # ("album", album["name"]),
            ("label", album["label"]),
            ("year", str(date.year)),
        ]:
            setattr(track, f"_{attribute_name.title()}", attribute)


def convert_to_datetime(arg: str) -> Union[datetime, str]:
    """Convert string to datetime.

    Args:
        arg: datetime as a YYYY-MM-DD string

    Returns:
        Either the datetime object to filter after or a string indicating that
            the most recent upload should be filtered.
    """
    if arg in ["most-recent", "all"]:
        return arg

    try:
        return datetime.strptime(arg, "%Y-%m-%d")
    except ValueError as exc:
        raise ValueError(
            f"Couldn't convert --date-filter argument '{arg}' into a datetime."
        ) from exc


def filter_tracks(
    tracks: Dict[str, Track], date_filter: Union[datetime, str]
) -> Dict[str, Track]:
    """Filter tracks using a date filter.

    Args:
        tracks: Dictionary of tracks to filter.
        date_filter: datetime to filter tracks with.

    Returns:
        Dictionary of filtered tracks.
    """
    if date_filter == "all":
        return tracks

    sorted_tracks = sorted(
        tracks.values(), key=lambda x: x.get_date_added(), reverse=True
    )

    if date_filter == "most-recent":
        date_filter = sorted_tracks[0].get_date_added()

    filtered_tracks = []
    for track in sorted_tracks:
        if track.get_date_added() < date_filter:
            break
        filtered_tracks.append(track)

    return {track.get_id(): track for track in filtered_tracks}


if __name__ == "__main__":
    arg_parser = ArgumentParser()
    arg_parser.add_argument("--collection", help="Path to a collection.")
    arg_parser.add_argument("--config", help="Path to a config.yaml.")
    arg_parser.add_argument(
        "--date-filter",
        type=convert_to_datetime,
        default="most-recent",
        help=(
            "Datetime to after which tracks should have tags added from "
            "Spotify. Default is 'most-recent'."
        ),
    )
    arg_parser.add_argument(
        "--mode",
        choices=["bulk", "interactive"],
        default="interactive",
        help="Bulk process tracks or review individual tracks.",
    )
    arg_parser.add_argument("--output", help="Path to output collection.")
    arg_parser.add_argument(
        "--query-limit",
        type=int,
        default=50,
        help="Number of Spotify query results.",
    )
    arg_parser.add_argument(
        "--similarity",
        type=int,
        default=90,
        help="Minimum Levenshtein distance for a Spotify query result to match a track.",
    )
    args = arg_parser.parse_args()

    # Build config, instantiate collection, and get tracks.
    config = build_config(args.config)
    collection = PLATFORM_REGISTRY[config.collection.platform]["collection"](
        path=args.collection or config.collection.collection_path
    )
    tracks = filter_tracks(collection.get_tracks(), args.date_filter)
    spotify = get_spotify_client(config)
    playlist_class = PLATFORM_REGISTRY[config.collection.platform]["playlist"]

    # Add tags from Spotify.
    if args.mode == "bulk":
        with (
            tqdm(total=len(tracks), desc="Adding tags from Spotify") as pbar,
            ThreadPoolExecutor(max_workers=os.cpu_count() * 4) as pool,
        ):
            futures = [
                pool.submit(
                    get_spotify_tags_thread,
                    track,
                    spotify,
                    args.similarity,
                    args.query_limit,
                )
                for track in tracks.values()
            ]
            for future in as_completed(futures):
                future.result()
                pbar.update(1)

        playlist = playlist_class.new_playlist(name="Matches", tracks=tracks)
        collection.add_playlist(playlist)
    elif args.mode == "interactive":
        no_matches = {}
        same_tracks = {}
        updated_tracks = {}

        for track_id, track in tqdm(tracks.items()):
            # Query Spotify for matching tracks
            title = track._Name
            artist = track.get_artists()
            query = f"track:{title} artist:{artist}"
            results = spotify.search(
                q=query, type="track", limit=args.query_limit
            )

            # Get result most similar to query.
            result, _ = filter_results(
                spotify, results, args.similarity, title, artist
            )

            # Mark track as a non-match and continue.
            if not result:
                no_matches[track_id] = track
                continue

            # Query album for the track.
            album = spotify.album(result["album"]["id"])
            for date_format in ["%Y-%m-%d", "%Y-%m", "%Y"]:
                try:
                    date = datetime.strptime(
                        album["release_date"], date_format
                    )
                except ValueError:
                    continue

            # Display comparison of track data from Collection vs. Spotify.
            artists = ", ".join([x["name"] for x in result["artists"]])
            print(
                f"C: {title} - {artist}: "
                f"{track._Album}, {track._Label}, {track._Year}\n"
                f"S: {result['name']} - {artists}: "
                f"{album['name']}, {album['label']}, {str(date.year)}"
            )

            # Check input for signal to skip or accept new tags.
            next_track = False
            while not next_track:
                resp = input("[y] accept, [n] skip")
                if "n" in resp:
                    same_tracks[track_id] = track
                    next_track = True
                elif "y" in resp:
                    for attribute_name, attribute in [
                        ("album", album["name"]),
                        ("label", album["label"]),
                        ("year", str(date.year)),
                    ]:
                        setattr(track, f"_{attribute_name.title()}", attribute)
                        updated_tracks[track_id] = track
                    next_track = True
                else:
                    print(
                        "Response must contain either 'n' or 'y' -- try again!"
                    )

        playlist_class = PLATFORM_REGISTRY[config.collection.platform][
            "playlist"
        ]
        for name, set_tracks in [
            ("No matches", no_matches),
            ("Same", same_tracks),
            ("Updated", updated_tracks),
        ]:
            playlist = playlist_class.new_playlist(
                name=name, tracks=set_tracks
            )
            collection.add_playlist(playlist)

    collection.serialize(path=args.output or config.collection.collection_path)



"""This is a script for moving tag data around for consistent formatting.

Supported functions include:

* moving particular tags from the `Comments` field to the `Genre` field
* moving particular tags from in the `Comments` field to the front of that field
"""

from argparse import ArgumentParser
from pathlib import Path
import re
from typing import Any, Dict, Set

import bs4
from bs4 import BeautifulSoup


def parse_args() -> Dict[str, Any]:
    """Parses command-line arguments.

    Returns:
        Dictionary of arguments.
    """

    def parse_path(path):
        try:
            path = Path(path)
            if not path.exists():
                raise FileNotFoundError(f"{path} does not exist.")
        except Exception as exc:
            raise TypeError(f"{path} is not a valid path.") from exc
        return path

    parser = ArgumentParser()
    parser.add_argument(
        "--mode",
        "-m",
        type=str,
        choices=[
            "move-genres-from-comments",
            "move-tags-to-front-of-comments",
        ],
        required=True,
        help="Path to Rekordbox XML file.",
    )
    parser.add_argument(
        "--xml",
        "-x",
        type=parse_path,
        required=True,
        help="Path to Rekordbox XML file.",
    )
    parser.add_argument(
        "--output",
        "-o",
        type=str,
        default="output_rekordbox.xml",
        help="Path to output Rekordbox XML file.",
    )
    parser.add_argument(
        "--tags",
        "-t",
        type=str,
        nargs="+",
        required=True,
        help='List of "My Tags" to be relocated.',
    )
    parser.add_argument(
        "--tag_delimiter",
        "-g",
        type=str,
        default="/",
        help="Character used to separate tags.",
    )

    return vars(parser.parse_args())


def move_genres_from_comments(
    track: bs4.element.Tag,
    my_tags: Set[str],
    tags: Set[str],
    tag_delimiter: str,
) -> bs4.element.Tag:
    """Moves `tags` from `my_tags` to the `Genre` field.

    Args:
        track: Track node.
        my_tags: Tags extracted from the `Comments` field.
        tags: Tags targeted for relocation.
        tag_delimiter: Delimiter used to separate tags within a field.

    Returns:
        Updated track node.
    """
    # Filter "My Tags" for tags belonging to the provided set of tags
    # to be relocated.
    new_genre_tags = []
    new_comments_tags = []
    for tag in my_tags.group().split("/"):
        tag = tag.strip()
        if not tag in tags:
            new_comments_tags.append(tag)
            continue
        new_genre_tags.append(tag)
    if not new_genre_tags:
        return track
    # Get pre-existing tags and split them on the chosen delimiter.
    current_tags = [x.strip() for x in track["Genre"].split(tag_delimiter)]
    # Deduplicate tags across pre-existing tags and "My Tags".
    if current_tags:
        new_genre_tags = set(current_tags).union(set(new_genre_tags))
    # Write new tags to field.
    track["Genre"] = f" {tag_delimiter} ".join(new_genre_tags)
    track["Comments"] = " / ".join(new_comments_tags)

    return track


def move_tags_to_front_of_comments(
    track: bs4.element.Tag,
    my_tags: Set[str],
    tags: Set[str],
    **kwargs,  # pylint: disable=unused-argument
) -> bs4.element.Tag:
    """Moves `tags` from `my_tags` to the front of the `Comments` field.

    Args:
        track: Track node.
        my_tags: Tags extracted from the `Comments` field.
        tags: Tags targeted for relocation.

    Returns:
        Updated track node.
    """
    tags_to_move = []
    remainder_tags = []
    for tag in my_tags.group().split("/"):
        tag = tag.strip()
        if tag in tags:
            tags_to_move.append(tag)
        else:
            remainder_tags.append(tag)
    if not tags_to_move:
        return track
    if len(tags_to_move) > 1:
        print(f"{track['Name']} has more than 1 situation tag: {tags_to_move}")
    comment_prefix = track["Comments"].split("/* ")[0]
    comment_suffix = track["Comments"].split(" */")[-1]
    new_my_tags = " / ".join(tags_to_move + remainder_tags)
    # Write new tags to Comments field.
    track["Comments"] = f"{comment_prefix} /* {new_my_tags} */{comment_suffix}"

    return track


move_functions = {
    "move-genres-from-comments": move_genres_from_comments,
    "move-tags-to-front-of-comments": move_tags_to_front_of_comments,
}


def move_tags(
    mode: str,
    xml: str,
    tags: Set[str],
    output: str,
    tag_delimiter: str,
):
    """Moved data from the `Comments` field.

    Behavior depends on the `mode` argument.

    The provided XML must be an valid XML export:
        `File > Export Collection in xml format`

    The provided tags must match exactly the data written to the Comments field
    (case-sensitive).

    The provided output must be the path to a XML file from which you'll import
    the adjusted data.

    Comments field tags are expected to have the format `/* tag1 / tag2 */`
    which is the way Rekordbox's "My Tag" system writes the data to the
    Comments field.

    Rekordbox's "My Tag" system will write data to the Comments field only if
    the following setting is enabled:
        `Preferences > Advanced > Browse > Add "My Tag" to the "Comments"`

    Args:
        xml: Path to a Rekordbox XML file.
        tags: Set of "My Tags" to be relocated.
        output: Path to an output Rekordbox XML file.
        tag_delimiter: Character used to delimit field tags.

    Raises:
        RuntimeError: `xml` must be a valid XML file.
    """
    try:
        with open(xml, mode="r", encoding="utf-8") as _file:
            database = BeautifulSoup(_file.read(), "xml")
    except Exception as exc:
        raise RuntimeError(
            "Are you sure the provided XML is valid? Parsing it failed with "
            f"the following exception:\n{exc}"
        ) from exc

    # Regular expression to isolate "My Tags".
    my_tag_regex = re.compile(r"(?<=\/\*).*(?=\*\/)")

    for track in database.find_all("TRACK"):
        # Make sure this isn't a TRACK node for a playlist.
        if not track.get("Location"):
            continue
        # Get "My Tags", if any, from the Comments field.
        my_tags = re.search(my_tag_regex, track["Comments"])
        if not my_tags:
            continue
        track = move_functions[mode](
            track=track,
            my_tags=my_tags,
            tags=tags,
            tag_delimiter=tag_delimiter,
        )

    # Write output rekordbox.xml to file.
    with open(output, mode="wb", encoding=database.original_encoding) as _file:
        _file.write(database.prettify("utf-8"))


if __name__ == "__main__":
    parsed_args = parse_args()
    if not parsed_args["tags"]:
        raise RuntimeError(
            'The --tags option must include a list of "My Tag" to be relocated'
        )
    parsed_args["tags"] = set(parsed_args["tags"])
    move_tags(**parsed_args)



"""Testing for the helpers module."""

import inspect
import logging
import re
from pathlib import Path
from unittest import mock

import pytest

try:
    import tomllib
except ImportError:
    import tomli as tomllib

from djtools.configs.config import LogLevel
from djtools.configs.helpers import _arg_parse, build_config, ConfigLoadFailure
from djtools.version import get_version

from ..test_utils import mock_exists, MockOpen


@pytest.mark.parametrize(
    "log_level_name, log_level", [("DEBUG", 10), ("WARNING", 30)]
)
@mock.patch("argparse.ArgumentParser.parse_args")
def test_arg_parse_sets_log_level(
    mock_parse_args, namespace, log_level_name, log_level
):
    """Test for the arg_parse function."""
    namespace.log_level = log_level_name
    mock_parse_args.return_value = namespace
    _arg_parse()
    logger = logging.getLogger("djtools.configs.helpers")
    assert logger.getEffectiveLevel() == log_level


@mock.patch("argparse.ArgumentParser.parse_args")
def test_arg_parse_gets_version(mock_parse_args, namespace, capsys):
    """Test for the arg_parse function."""
    namespace.version = True
    mock_parse_args.return_value = namespace
    with pytest.raises(SystemExit):
        _arg_parse()
    with open(
        Path(__file__).parent.parent.parent / "pyproject.toml", mode="rb"
    ) as _file:
        toml_dict = tomllib.load(_file)
    assert (
        capsys.readouterr().out.replace(".", "").strip()
        == toml_dict["project"]["version"].replace(".", "").strip()
    )


@mock.patch("argparse.ArgumentParser.parse_args")
def test_arg_parse_links_configs_dir_does_exist(
    mock_parse_args, tmpdir, namespace
):
    """Test for the arg_parse function."""
    config_path = Path(tmpdir) / "new_dir" / "link_dir"
    namespace.link_configs = config_path
    mock_parse_args.return_value = namespace
    assert not config_path.parent.exists()
    _arg_parse()
    assert config_path.exists()
    assert config_path.is_symlink()


@mock.patch("argparse.ArgumentParser.parse_args")
def test_arg_parse_links_configs_dir_parent_does_not_exist(
    mock_parse_args, tmpdir, namespace
):
    """Test for the arg_parse function."""
    config_path = str(tmpdir)
    namespace.link_configs = config_path
    mock_parse_args.return_value = namespace
    with pytest.raises(
        ValueError,
        # NOTE(a-rich): WindowsPath needs to be escaped for `\` characters to
        # appear in the `match` argument.
        match=re.escape(
            f"{config_path} must be a directory that does not already exist"
        ),
    ):
        _arg_parse()


@mock.patch(
    "builtins.open",
    MockOpen(
        files=["config.yaml"],
        content="yaml_valid: {false",
    ).open,
)
def test_build_config_invalid_config_yaml(caplog):
    """Test for the build_config function."""
    caplog.set_level("CRITICAL")
    with (
        mock.patch.object(Path, "exists", return_value=True),
        pytest.raises(ConfigLoadFailure),
    ):
        build_config()
    assert "Error reading" in caplog.records[0].message


@mock.patch("djtools.spotify.helpers.get_spotify_client", mock.Mock())
@mock.patch("argparse.ArgumentParser.parse_args")
def test_build_config_no_config_yaml(
    mock_parse_args, namespace, config_file_teardown
):
    """Test for the build_config function."""
    # pylint: disable=unused-argument
    mock_parse_args.return_value = namespace
    config_dir = (
        Path(__file__).parent.parent.parent / "src" / "djtools" / "configs"
    )
    config_file = config_dir / "config.yaml"
    assert not config_file.exists()
    build_config()
    assert config_file.exists()


@mock.patch(
    "builtins.open",
    MockOpen(
        files=["config.yaml"],
        content="configs:\n  log_level: INFO",
    ).open,
)
@mock.patch(
    "djtools.configs.helpers.Path.exists",
    lambda path: mock_exists(
        [
            ("config.yaml", True),
        ],
        path,
    ),
)
@mock.patch("argparse.ArgumentParser.parse_args")
def test_build_config_overrides_using_args(mock_parse_args, namespace):
    """Test for the build_config function."""
    namespace.log_level = LogLevel.CRITICAL.value
    mock_parse_args.return_value = namespace
    config = build_config()
    assert config.log_level == LogLevel.CRITICAL


@mock.patch("argparse.ArgumentParser.parse_args")
def test_build_config_reads_generated_config(
    mock_parse_args, namespace, config_file_teardown
):
    """Test for the build_config function."""
    # pylint: disable=unused-argument
    namespace.log_level = "INFO"
    mock_parse_args.return_value = namespace
    config_dir = (
        Path(__file__).parent.parent.parent / "src" / "djtools" / "configs"
    )
    config_file = config_dir / "config.yaml"

    # config.yaml won't exist so build_config generates one.
    assert not config_file.exists()
    first_config = build_config()

    # config.yaml now exists so build_config reads the pre-existing one.
    assert config_file.exists()
    second_config = build_config()

    # Both configs should be identical.
    assert first_config == second_config


@mock.patch("argparse.ArgumentParser.parse_args")
def test_build_config_version(
    mock_parse_args,
    namespace,
    capsys,
    config_file_teardown,
):
    """Test for the build_config function."""
    # pylint: disable=unused-argument
    namespace.version = True
    mock_parse_args.return_value = namespace
    with pytest.raises(SystemExit):
        build_config()
    assert capsys.readouterr().out == f"{get_version()}\n"


def test_stack_inspection_for_cli_arg_parsing():
    """Test stack inspection for cli arg parsing."""
    stack = inspect.stack()
    entry_frame = stack[-1]
    test_loc = str(Path("bin") / "pytest")
    windows_loc = str(Path("lib") / "runpy.py")
    windows_frozen = "<frozen runpy>"
    assert entry_frame[1].endswith((test_loc, windows_loc, windows_frozen))



"""Testing for the copy_playlists module."""

from pathlib import Path

import pytest

from djtools.collection.copy_playlists import copy_playlists
from djtools.collection.rekordbox_collection import RekordboxCollection


def test_copy_playlists_makes_destination_folder(
    tmpdir, config, rekordbox_xml
):
    """Test for the copy_playlists function."""
    target_playlists = ["Hip Hop"]
    test_output_dir = Path(tmpdir) / "output"
    config.collection.collection_path = rekordbox_xml
    config.collection.copy_playlists = target_playlists
    config.collection.copy_playlists_destination = Path(test_output_dir)
    new_collection = tmpdir / "test_collection"
    assert not config.collection.copy_playlists_destination.exists()
    copy_playlists(config, path=new_collection)
    assert config.collection.copy_playlists_destination.exists()


def test_copy_playlists_handles_invalid_playlist(
    tmpdir, config, rekordbox_xml
):
    """Test for the copy_playlists function."""
    playlist = "invalid_playlist"
    config.collection.collection_path = rekordbox_xml
    config.collection.copy_playlists = [playlist]
    config.collection.copy_playlists_destination = Path(tmpdir)
    with pytest.raises(LookupError, match=f"{playlist} not found"):
        copy_playlists(config)


def test_copy_playlists_find_multiple_playlists_with_the_same_name(
    tmpdir, config, rekordbox_collection, rekordbox_xml
):
    """Test for the copy_playlists function."""
    target_playlist = "Dark"
    config.collection.collection_path = rekordbox_xml
    config.collection.copy_playlists = [target_playlist]
    config.collection.copy_playlists_destination = Path(tmpdir)
    new_collection = tmpdir / "test_collection"
    old_playlist_count = len(
        rekordbox_collection.get_playlists(target_playlist)
    )
    copy_playlists(config, path=new_collection)
    collection = RekordboxCollection(new_collection)
    new_playlist_count = len(collection.get_playlists(target_playlist))
    assert old_playlist_count == new_playlist_count


def test_copy_playlists_copies_files(
    tmpdir, config, rekordbox_collection, rekordbox_xml
):
    """Test for the copy_playlists function."""
    target_playlists = ["Hip Hop", "Dark"]
    test_output_dir = Path(tmpdir) / "output"
    config.collection.collection_path = rekordbox_xml
    config.collection.copy_playlists = target_playlists
    config.collection.copy_playlists_destination = test_output_dir
    new_collection = Path(tmpdir) / "test_collection"
    old_tracks = {
        track_id: track
        for target_playlist in target_playlists
        for playlist in rekordbox_collection.get_playlists(target_playlist)
        for track_id, track in playlist.get_tracks().items()
    }
    copy_playlists(config, path=new_collection)
    num_copied_tracks = len(list(test_output_dir.iterdir()))
    assert num_copied_tracks == len(old_tracks)
    collection = RekordboxCollection(new_collection)
    new_tracks = {
        track_id: track
        for target_playlist in target_playlists
        for playlist in collection.get_playlists(target_playlist)
        for track_id, track in playlist.get_tracks().items()
    }
    for track_id, track in old_tracks.items():
        old_loc = track.get_location()
        new_loc = new_tracks[track_id].get_location()
        assert old_loc != new_loc
        assert old_loc.name == new_loc.name
        assert new_loc.parent == test_output_dir


def test_copy_playlists_creates_new_collection(config, rekordbox_xml, tmpdir):
    """Test for the copy_playlists function."""
    playlist = "Hip Hop"
    config.collection.collection_path = rekordbox_xml
    config.collection.copy_playlists = [playlist]
    config.collection.copy_playlists_destination = Path(tmpdir)
    new_collection = tmpdir / "test_collection"
    assert not new_collection.exists()
    copy_playlists(config, path=new_collection)
    assert new_collection.exists()


def test_copy_playlists_creates_new_collection_with_default_path(
    config, rekordbox_xml, tmpdir
):
    """Test for the copy_playlists function."""
    playlist = "Hip Hop"
    config.collection.collection_path = rekordbox_xml
    config.collection.copy_playlists = [playlist]
    config.collection.copy_playlists_destination = Path(tmpdir)
    new_collection = (
        config.collection.copy_playlists_destination
        / f"copied_playlists_collection{rekordbox_xml.suffix}"
    )
    assert not new_collection.exists()
    copy_playlists(config)
    assert new_collection.exists()



"""Testing for the config module."""

from unittest import mock

import pytest

from djtools.spotify.config import SpotifyConfig


@mock.patch("djtools.spotify.helpers.Client")
def test_baseconfig_invalid_spotify_credentials(mock_client):
    """Test for the SpotifyConfig class with invalid credentials."""
    mock_client.return_value.current_user.side_effect = Exception()
    cfg = {
        "spotify_client_id": "not a real ID",
        "spotify_client_secret": "not a real secret",
        "spotify_playlist_from_upload": True,
        "spotify_redirect_uri": "not a real URI",
        "spotify_username": "not a real username",
    }
    with pytest.raises(RuntimeError, match="Spotify credentials are invalid!"):
        SpotifyConfig(**cfg)


def test_spotifyconfig_no_spotify_credentials():
    """Test for the SpotifyConfig class without credentials."""
    cfg = {"spotify_playlist_from_upload": True, "spotify_client_id": ""}
    with pytest.raises(
        RuntimeError,
        match=(
            "Without all the configuration options spotify_client_id, "
            "spotify_client_secret, spotify_redirect_uri, and "
            "spotify_username set to valid values, you cannot use "
            "spotify_playlists or spotify_playlist_from_upload"
        ),
    ):
        SpotifyConfig(**cfg)


@mock.patch("djtools.spotify.helpers.Client")
def test_spotifyconfig_no_reddit_credentials(_mock_client):
    """Test for the SpotifyConfig class without Reddit credentials."""
    cfg = {
        "reddit_client_id": "",
        "spotify_client_id": "id",
        "spotify_client_secret": "secret",
        "spotify_playlists": True,
        "spotify_redirect_uri": "uri",
        "spotify_username": "name",
    }
    with pytest.raises(
        RuntimeError,
        match=(
            "Without all the configuration options reddit_client_id, "
            "reddit_client_secret, and reddit_user_agent, set to valid "
            "values, you cannot use spotify_playlists"
        ),
    ):
        SpotifyConfig(**cfg)



"""Script that removes files from a USB if they're not present in a collection."""

from argparse import ArgumentParser, Namespace
from pathlib import Path

from djtools.configs import build_config
from djtools.collection.platform_registry import PLATFORM_REGISTRY


# pylint: disable=missing-function-docstring


def parse_args() -> Namespace:
    parser = ArgumentParser()
    parser.add_argument("--config")
    parser.add_argument("--usb")
    args = parser.parse_args()

    return args


def strip_base_path(path: str, spliter: str = "DJ Music") -> str:
    return path.split(spliter)[-1]


def main():
    args = parse_args()

    config_path = Path(args.config)
    usb_path = Path(args.usb)

    config = build_config(config_path)
    collection = PLATFORM_REGISTRY[config.platform]["collection"](
        path=config.collection_path
    )

    tracks = {
        strip_base_path(track.get_location().as_posix()): track
        for _, track in collection.get_tracks().items()
    }
    files = {
        strip_base_path(path.as_posix()): path
        for path in usb_path.rglob("DJ Music/**/*.*")
    }
    files_not_in_collection = {
        path: path_obj
        for path, path_obj in files.items()
        if path not in tracks
    }

    print(f"Found {len(files_not_in_collection)} files not in collection.")

    for path in files_not_in_collection.values():
        path.unlink()


if __name__ == "__main__":
    main()



"""Testing for the collection module."""

import bs4

from djtools.collection.rekordbox_collection import (
    CustomSubstitution,
    RekordboxCollection,
    UnsortedAttributes,
)
from djtools.collection.rekordbox_playlist import RekordboxPlaylist
from djtools.collection.rekordbox_track import RekordboxTrack


def test_rekordboxcollection_add_playlist(rekordbox_xml):
    """Test RekordboxCollection class."""
    collection = RekordboxCollection(path=rekordbox_xml)
    num_playlists = len(collection.get_playlists())

    # The test playlist doesn't exist yet.
    test_playlist_name = "TEST"
    assert collection.get_playlists(test_playlist_name) == []

    # The test playlist does exist after inserting it.
    test_playlist = RekordboxPlaylist.new_playlist(
        test_playlist_name, tracks={}
    )
    collection.add_playlist(test_playlist)
    assert collection.get_playlists(test_playlist_name) == [test_playlist]

    # The total playlist count has increased by one.
    assert len(collection.get_playlists()) == num_playlists + 1


def test_rekordboxcollection_get_all_tags(rekordbox_collection):
    """Test RekordboxCollection class."""
    # Manually build the sets of genre tags and other tags from the tracks in
    # the collection.
    tracks = rekordbox_collection.get_tracks().values()
    genre_tags, all_tags = set(), set()
    for track in tracks:
        genre_tags.update(track.get_genre_tags())
        all_tags.update(set(track.get_tags()))
    other_tags = all_tags.difference(genre_tags)

    # Compare the manually constructed tag sets with tag sets returned from the
    # get_all_tags methods.
    expected = {
        "genres": sorted(genre_tags),
        "other": sorted(other_tags),
    }
    tags = rekordbox_collection.get_all_tags()
    assert tags == expected


def test_rekordboxcollection_get_playlists(rekordbox_xml):
    """Test RekordboxCollection class."""
    collection = RekordboxCollection(path=rekordbox_xml)

    # Without being provided any playlists to get, the get_playlists method
    # returns the ROOT playlist.
    assert collection.get_playlists().get_name() == "ROOT"

    # When provided a playlist name, a list containing all the playlists with
    # matching names is returned.
    dark_playlists = collection.get_playlists("Dark")
    for playlist in dark_playlists:
        assert playlist.get_name() == "Dark"


def test_rekordboxcollection_get_tracks(
    rekordbox_xml, rekordbox_collection_tag
):
    """Test RekordboxCollection class."""
    collection = RekordboxCollection(path=rekordbox_xml)

    # Manually construct the track dictionary stored when deserializing a
    # collection.
    tracks = {
        track["TrackID"]: RekordboxTrack(track)
        for track in rekordbox_collection_tag.find_all("TRACK")
        if track.get("Location")
    }

    # Compare the manually constructed track dictionary with the one internal
    # to the deserialized collection.
    assert (
        a[0] == b[0] and str(a[1]) == str(b[1])
        for a, b in zip(tracks.items(), collection.get_tracks().items())
    )


def test_rekordboxcollection_serialization(rekordbox_xml):
    """Test RekordboxCollection class."""
    collection = RekordboxCollection(path=rekordbox_xml)

    assert (
        repr(collection)
        == f"""RekordboxCollection(
    path="{rekordbox_xml}",
    tracks=4,
    playlists=3,
)"""
    )

    # Serializing a collection returns the path to the serialized collection
    # which will overwrite the path from which the collection was deserialized
    # from...
    assert collection.serialize() == rekordbox_xml

    # ...unless a different path is provided explicitly.
    new_path = rekordbox_xml.parent / "test_collection"
    serialized_collection = collection.serialize(path=new_path)
    assert new_path.exists()
    assert rekordbox_xml != serialized_collection

    # Serialization / deserialization is a symmetrical operation.
    try:
        RekordboxCollection.validate(rekordbox_xml, serialized_collection)
    except AssertionError:
        assert False, "RekordboxCollection validation failed!"


def test_rekordboxcollection_set_tracks(rekordbox_xml):
    """Test RekordboxCollection class."""
    collection = RekordboxCollection(path=rekordbox_xml)

    # Get the original set of tracks.
    original_tracks = collection.get_tracks()

    # Create a new dictionary of tracks which contains only one (key, value)
    # pair.
    new_tracks = {}
    for key, value in original_tracks.items():
        new_tracks[key] = value
        break

    # Set the collections tracks to this new subset of tracks.
    collection.set_tracks(new_tracks)

    # The new tracks are different from the original tracks.
    assert new_tracks != original_tracks

    # The collection now returns the new tracks.
    assert collection.get_tracks() == new_tracks


def test_customsubstitution():
    """Test CustomSubstitution class."""
    test_string = """Bob's cat is "cute" & <furry>"""
    expected = "Bob&apos;s cat is &quot;cute&quot; &amp; &lt;furry&gt;"
    result = CustomSubstitution.substitute_xml(test_string)
    assert result == expected


def test_unsortedattributes_formatter():
    """Test UnsortedAttributes class."""
    expected = '<NODE Z_attr="" A_attr="">\n</NODE>\n'
    a_first_tag = bs4.Tag(name="NODE", attrs={"A_attr": "", "Z_attr": ""})
    z_first_tag = bs4.Tag(name="NODE", attrs={"Z_attr": "", "A_attr": ""})
    # BeautifulSoup alphabetizes attributes by default.
    assert str(z_first_tag) == str(a_first_tag)
    # Using UnsortedAttributes formatter will cause the Tag to render the
    # attributes in their original order.
    assert z_first_tag.prettify(formatter=UnsortedAttributes()) == expected



"""Testing for the playlist_filters module."""

from unittest import mock

import pytest

from djtools.collection.playlist_filters import (
    ComplexTrackFilter,
    HipHopFilter,
    MinimalDeepTechFilter,
    PlaylistFilter,
    TransitionTrackFilter,
)
from djtools.collection.rekordbox_playlist import RekordboxPlaylist
from djtools.collection.rekordbox_track import RekordboxTrack


def test_playlistfilter_cannot_be_instantiated():
    """Test for the PlaylistFilter class."""
    with pytest.raises(
        TypeError,
        match="Can't instantiate abstract class PlaylistFilter",
    ):
        PlaylistFilter()


@pytest.mark.parametrize(
    "parent_playlist,playlist,hip_hop_playlist,bass_hip_hop_playlist",
    [
        ("Bass", "Hip Hop", True, True),
        ("Not Bass", "Hip Hop", True, False),
        ("something", "irrelevant", False, False),
    ],
)
def test_hiphopfilter_detects_playlists(
    parent_playlist,
    playlist,
    hip_hop_playlist,
    bass_hip_hop_playlist,
):
    """Test for the HipHopFilter class."""
    track_filter = HipHopFilter()
    playlist = RekordboxPlaylist.new_playlist(playlist, tracks={})
    parent_playlist = RekordboxPlaylist.new_playlist(
        parent_playlist, playlists=[playlist]
    )
    playlist.set_parent(parent_playlist)
    result = track_filter.is_filter_playlist(playlist)
    assert result == hip_hop_playlist
    assert (
        track_filter._bass_hip_hop  # pylint: disable=protected-access
        == bass_hip_hop_playlist
    )


@pytest.mark.parametrize(
    "bass_hip_hop,tags,expected",
    [
        (True, ["Hip Hop", "R&B"], False),
        (True, ["Hip Hop", "Trap"], True),
        (False, ["Hip Hop", "R&B"], True),
        (False, ["Hip Hop", "Trap"], False),
    ],
)
def test_hiphopfilter_filters_tracks(
    bass_hip_hop, tags, expected, rekordbox_track
):
    """Test for the HipHopFilter class."""
    track_filter = HipHopFilter()
    with (
        mock.patch.object(
            track_filter, "_bass_hip_hop", bass_hip_hop, create=True
        ),
        mock.patch.object(RekordboxTrack, "get_genre_tags", lambda x: tags),
    ):
        result = track_filter.filter_track(rekordbox_track)
        assert result == expected


@pytest.mark.parametrize(
    "parent_playlist,playlist,minimal_deep_tech_playlist",
    [
        ("Techno", "Minimal Deep Tech", True),
        ("House", "Minimal Deep Tech", True),
        ("Techno", "Something", False),
        ("House", "Something", False),
        ("Something", "Minimal Deep Tech", False),
    ],
)
def test_minimaldeeptechfilter_detects_playlists(
    parent_playlist,
    playlist,
    minimal_deep_tech_playlist,
):
    """Test for the MinimalDeepTechFilter class."""
    track_filter = MinimalDeepTechFilter()
    playlist = RekordboxPlaylist.new_playlist(playlist, tracks={})
    parent_playlist = RekordboxPlaylist.new_playlist(
        parent_playlist, playlists=[playlist]
    )
    playlist.set_parent(parent_playlist)
    result = track_filter.is_filter_playlist(playlist)
    techno_playlist = (
        parent_playlist.get_name() == "Techno" and minimal_deep_tech_playlist
    )
    house_playlist = (
        parent_playlist.get_name() == "House" and minimal_deep_tech_playlist
    )
    assert result == minimal_deep_tech_playlist
    assert (
        track_filter._techno  # pylint: disable=protected-access
        == techno_playlist
    )
    assert (
        track_filter._house  # pylint: disable=protected-access
        == house_playlist
    )


@pytest.mark.parametrize(
    "techno,tags,expected",
    [
        (True, ["Techno", "Minimal Deep Tech"], True),
        (True, ["House", "Minimal Deep Tech"], False),
        (False, ["Techno", "Minimal Deep Tech"], False),
        (False, ["House", "Minimal Deep Tech"], True),
    ],
)
def test_minimaldeeptechfilter_filters_tracks(
    techno, tags, expected, rekordbox_track
):
    """Test for the MinimalDeepTechFilter class."""
    track_filter = MinimalDeepTechFilter()
    with (
        mock.patch.object(track_filter, "_techno", techno, create=True),
        mock.patch.object(track_filter, "_house", not techno, create=True),
        mock.patch.object(RekordboxTrack, "get_genre_tags", lambda x: tags),
    ):
        result = track_filter.filter_track(rekordbox_track)
        assert result == expected


@pytest.mark.parametrize(
    "parent_playlist,playlist,expected",
    [
        ("something", "complex", True),
        ("complex", "irrelevant", True),
        ("irrelevant", "something", False),
    ],
)
def test_complextrackfilter_detects_playlists(
    parent_playlist,
    playlist,
    expected,
):
    """Test for the ComplexTrackFilter class."""
    track_filter = ComplexTrackFilter()
    playlist = RekordboxPlaylist.new_playlist(playlist, tracks={})
    parent_playlist = RekordboxPlaylist.new_playlist(
        parent_playlist, playlists=[playlist]
    )
    playlist.set_parent(parent_playlist)
    result = track_filter.is_filter_playlist(playlist)
    assert result == expected


@pytest.mark.parametrize(
    "max_tags,tags,expected",
    [
        (1, set(["Tag"]), True),
        (2, set(["Tag"]), False),
        (2, set(["Tag", "Another Tag"]), True),
        (3, set(["Tag", "Another Tag"]), False),
        (3, set(["Tag", "Another Tag", "Last Tag"]), True),
    ],
)
def test_complextrackfilter_filters_tracks(
    max_tags, tags, expected, rekordbox_track
):
    """Test for the ComplexTrackFilter class."""
    track_filter = ComplexTrackFilter(max_tags)
    rekordbox_track._Tags = tags  # pylint: disable=protected-access
    playlist = RekordboxPlaylist.new_playlist(
        "genres", tracks={rekordbox_track.get_id(): rekordbox_track}
    )
    root_playlist = RekordboxPlaylist.new_playlist(
        "complex", playlists=[playlist]
    )
    playlist.set_parent(root_playlist)
    result = track_filter.filter_track(rekordbox_track)
    assert result == expected


@pytest.mark.parametrize(
    "parent_playlist,playlist,expected",
    [
        ("transitions", "genres", True),
        ("transitions", "tempos", True),
        ("irrelevant", "transitions", False),
        ("irrelevant", "genres", False),
        ("irrelevant", "tempos", False),
    ],
)
def test_transitiontrackfilter_detects_playlists(
    parent_playlist,
    playlist,
    expected,
):
    """Test for the TransitionTrackFilter class."""
    track_filter = TransitionTrackFilter()
    playlist = RekordboxPlaylist.new_playlist(playlist, tracks={})
    parent_playlist = RekordboxPlaylist.new_playlist(
        parent_playlist, playlists=[playlist]
    )
    playlist.set_parent(parent_playlist)
    result = track_filter.is_filter_playlist(playlist)
    assert result == expected


def test_transitiontrackfilter_handles_playlist_with_multiple_supported_types():
    """Test for the TransitionTrackFilter class."""
    track_filter = TransitionTrackFilter()
    bad_playlist = "genres & tempos"
    playlist = RekordboxPlaylist.new_playlist(bad_playlist, tracks={})
    parent_playlist = RekordboxPlaylist.new_playlist(
        "transitions", playlists=[playlist]
    )
    playlist.set_parent(parent_playlist)
    with pytest.raises(
        ValueError,
        match=(
            f'"{bad_playlist}" matches multiple playlist types: genre, '
            "tempo"
        ),
    ):
        track_filter.is_filter_playlist(playlist)


@pytest.mark.parametrize(
    "separator,comments,playlist_type,expected",
    [
        ("/", "[130 / 140]", "tempo", True),
        ("/", "[130 / 140]", "genre", False),
        ("/", "[Techno / Dubstep]", "genre", True),
        ("/", "[Techno / Dubstep]", "tempo", False),
    ],
)
def test_transitiontrackfilter_filters_tracks(
    separator, comments, playlist_type, expected, rekordbox_track
):
    """Test for the TransitionTrackFilter class."""
    track_filter = TransitionTrackFilter(separator)
    rekordbox_track._Comments = comments  # pylint: disable=protected-access
    playlist = RekordboxPlaylist.new_playlist(
        "genres", tracks={rekordbox_track.get_id(): rekordbox_track}
    )
    root_playlist = RekordboxPlaylist.new_playlist(
        "transitions", playlists=[playlist]
    )
    playlist.set_parent(root_playlist)
    with mock.patch.object(
        track_filter, "_playlist_type", playlist_type, create=True
    ):
        result = track_filter.filter_track(rekordbox_track)
    assert result == expected



"""Testing for the sync_operations module."""

from pathlib import Path
from unittest import mock

import pytest

from djtools.collection.config import RegisteredPlatforms
from djtools.sync.sync_operations import (
    download_collection,
    download_music,
    upload_collection,
    upload_music,
)


# pylint: disable=duplicate-code
TEST_BUCKET = "s3://some-bucket.com"


@pytest.mark.parametrize("playlist_name", ["", "playlist Uploads"])
@mock.patch(
    "djtools.sync.sync_operations.compare_tracks",
    mock.Mock(
        return_value=(
            [Path("file.mp3")],
            [Path("playlist/file.mp3")],
        ),
    ),
)
def test_download_music(playlist_name, config, tmpdir, caplog):
    """Test for the download_music function."""
    caplog.set_level("INFO")
    config.sync.bucket_url = TEST_BUCKET
    config.sync.usb_path = tmpdir
    config.sync.download_spotify_playlist = playlist_name
    write_path = Path(tmpdir) / "DJ Music" / "file.mp3"
    cmd = [
        "aws",
        "s3",
        "sync",
        f"{TEST_BUCKET}/dj/music/",
        write_path.parent.as_posix(),
    ]
    if playlist_name:
        cmd += ["--exclude", "*", "--include", "playlist/file.mp3"]
    cmd += ["--size-only"]

    def dummy_func():
        with open(write_path, mode="w", encoding="utf-8") as _file:
            _file.write("")

    with mock.patch(
        "djtools.sync.sync_operations.run_sync",
        side_effect=lambda *args, **kwargs: dummy_func(),
    ) as mock_run_sync:
        download_music(config)
        mock_run_sync.assert_called_with(cmd, TEST_BUCKET)
    assert caplog.records[0].message == "Downloading track collection..."
    assert (
        caplog.records[1].message == f"Found 0 files at {config.sync.usb_path}"
    )
    assert caplog.records[2].message == " ".join(cmd)
    assert caplog.records[3].message == "Found 1 new files"
    assert Path(caplog.records[4].message).name == "file.mp3"


@mock.patch("djtools.utils.helpers.get_spotify_client", mock.Mock())
def test_download_spotify_playlist_handles_no_matches(config, caplog):
    """Test for the download_music function."""
    caplog.set_level("WARNING")
    config.sync.download_spotify_playlist = "not-a-real-user Uploads"
    beatcloud_tracks = download_music(config)
    assert not beatcloud_tracks
    assert caplog.records[0].message == (
        "not-a-real-user Uploads not in spotify_playlists.yaml"
    )
    assert caplog.records[1].message == (
        "There are no Spotify tracks; make sure download_spotify_playlist is "
        "a key in spotify_playlists.yaml"
    )
    assert caplog.records[2].message == (
        "No Beatcloud matches were found! Make sure you've supplied the "
        "correct playlist name."
    )


@pytest.mark.parametrize("collection_is_dir", [True, False])
@pytest.mark.parametrize("user_is_import_user", [True, False])
@mock.patch("djtools.sync.sync_operations.rewrite_track_paths")
@mock.patch("djtools.sync.sync_operations.Popen")
def test_download_collection(
    mock_popen,
    mock_rewrite_track_paths,
    collection_is_dir,
    user_is_import_user,
    config,
    rekordbox_xml,
    caplog,
):
    """Test for the download_collection function."""
    caplog.set_level("INFO")
    test_user = "test_user"
    other_user = "other_user"
    import_user = test_user if user_is_import_user else other_user
    config.sync.bucket_url = TEST_BUCKET
    config.sync.user = test_user
    config.sync.import_user = import_user
    config.collection.collection_path = rekordbox_xml
    config.collection.platform = RegisteredPlatforms.REKORDBOX
    new_xml = rekordbox_xml.parent / f"{import_user}_rekordbox.xml"
    new_xml.write_text(
        rekordbox_xml.read_text(encoding="utf-8"), encoding="utf-8"
    )
    process = mock_popen.return_value.__enter__.return_value
    process.wait.return_value = 0
    cmd = [
        "aws",
        "s3",
        "cp",
        f"{TEST_BUCKET}/dj/collections/{import_user}/"
        f"{config.collection.platform.value}_collection",
        # NOTE(a-rich): since we could be passing a `rekordbox_xml` formatted
        # as a WindowsPath, the comparison needs to be made with `str(new_xml)`
        # (rather than `new_xml.as_posix()`).
        new_xml.as_posix(),
    ]
    if collection_is_dir:
        cmd.append("--recursive")
    with mock.patch(
        "djtools.sync.sync_operations.Path.is_dir",
        lambda path: collection_is_dir,
    ):
        download_collection(config)
    mock_popen.assert_called_with(cmd)
    assert caplog.records[0].message == (
        f"Downloading {config.sync.import_user}'s {config.collection.platform.value} collection..."
    )
    assert caplog.records[1].message == " ".join(cmd)
    if not user_is_import_user:
        mock_rewrite_track_paths.assert_called_once()


@pytest.mark.parametrize(
    "discord_url", ["", "https://discord.com/api/webhooks/some-id/"]
)
@mock.patch("djtools.sync.sync_operations.run_sync", return_value=None)
@mock.patch("djtools.sync.sync_operations.webhook", return_value=None)
def test_upload_music(
    mock_webhook, mock_run_sync, discord_url, tmpdir, config, caplog
):
    """Test for the upload_music function."""
    caplog.set_level("INFO")
    config.sync.bucket_url = TEST_BUCKET
    config.sync.usb_path = tmpdir
    config.sync.discord_url = discord_url
    test_dir = Path(tmpdir) / "DJ Music"
    test_dir.mkdir(parents=True, exist_ok=True)
    for file_name in [".test_file.mp3", "test_file.mp3"]:
        with open(test_dir / file_name, mode="w", encoding="utf-8") as file_:
            file_.write("")
    upload_music(config)
    cmd = [
        "aws",
        "s3",
        "sync",
        test_dir.as_posix(),
        f"{TEST_BUCKET}/dj/music/",
        "--size-only",
    ]
    assert caplog.records[0].message == "Removed 1 files..."
    assert Path(caplog.records[1].message).name == ".test_file.mp3"
    assert caplog.records[2].message == "Uploading track collection..."
    assert caplog.records[3].message == " ".join(cmd)
    assert len(list(test_dir.iterdir())) == 1
    mock_run_sync.assert_called_with(cmd, TEST_BUCKET)
    if discord_url:
        mock_webhook.assert_called_with(
            "https://discord.com/api/webhooks/some-id/", content=None
        )


@pytest.mark.parametrize("collection_is_dir", [True, False])
@mock.patch("djtools.sync.sync_operations.Popen")
def test_upload_collection(
    mock_popen, collection_is_dir, config, rekordbox_xml, caplog
):
    """Test for the upload_collection function."""
    caplog.set_level("INFO")
    user = "user"
    config.sync.bucket_url = TEST_BUCKET
    config.sync.user = user
    config.collection.collection_path = rekordbox_xml
    config.collection.platform = RegisteredPlatforms.REKORDBOX
    cmd = [
        "aws",
        "s3",
        "cp",
        config.collection.collection_path.as_posix(),
        f"{TEST_BUCKET}/dj/collections/{user}/{config.collection.platform.value}_collection",
    ]
    if collection_is_dir:
        cmd.append("--recursive")
    process = mock_popen.return_value.__enter__.return_value
    process.wait.return_value = 0
    with mock.patch(
        "djtools.sync.sync_operations.Path.is_dir",
        lambda path: collection_is_dir,
    ):
        upload_collection(config)
    assert caplog.records[0].message == (
        f"Uploading {user}'s {config.collection.platform.value} collection..."
    )
    assert caplog.records[1].message == " ".join(cmd)
    mock_popen.assert_called_with(cmd)



"""This module creates or updates Spotify playlists.

Playlists can be created from:
1. Discord webhook output from `djtools.sync.sync_operations.upload_music`
2. Subreddit posts (directly linked Spotify tracks or fuzzy-matched titles)
"""

import asyncio
import logging
from pathlib import Path
from typing import Type

import pyperclip
import yaml

from djtools.spotify.helpers import (
    filter_results,
    get_playlist_ids,
    get_reddit_client,
    get_spotify_client,
    get_subreddit_posts,
    populate_playlist,
    write_playlist_ids,
)


BaseConfig = Type["BaseConfig"]

# Silence PRAW, Spotify, and urllib3 loggers
for logger_name in ["asyncprawcore", "spotipy", "urllib3"]:
    logging.getLogger(logger_name).setLevel(logging.CRITICAL)
logging.getLogger("asyncio").setLevel(logging.WARNING)

logger = logging.getLogger(__name__)


async def async_spotify_playlists(config: BaseConfig):
    """Updates Spotify playlists from subreddit posts.

    Args:
        config: Configuration object.
    """
    spotify = get_spotify_client(config)
    reddit = get_reddit_client(config)
    playlist_ids = get_playlist_ids()

    # Load praw cache
    praw_cache = {}
    cache_file = Path(__file__).parent / ".praw.cache"
    if cache_file.exists():
        with open(cache_file, mode="r", encoding="utf-8") as _file:
            praw_cache = yaml.load(_file, Loader=yaml.FullLoader) or {}

    # Create async tasks for each subreddit
    tasks = [
        asyncio.create_task(
            get_subreddit_posts(spotify, reddit, subreddit, config, praw_cache)
        )
        for subreddit in config.spotify.spotify_playlist_subreddits
    ]

    # Process results as they complete
    for task in asyncio.as_completed(tasks):
        tracks, subreddit = await task
        playlist_ids = populate_playlist(
            playlist_name=subreddit.name,
            playlist_ids=playlist_ids,
            spotify_username=config.spotify.spotify_username,
            spotify=spotify,
            tracks=tracks,
            playlist_limit=subreddit.limit,
            verbosity=config.verbosity,
        )

    await reddit.close()

    # Save state
    write_playlist_ids(playlist_ids)
    with open(cache_file, mode="w", encoding="utf-8") as _file:
        yaml.dump(praw_cache, _file)


def spotify_playlist_from_upload(config: BaseConfig):
    """Generates a Spotify playlist from Discord webhook output.

    Uses clipboard contents containing upload_music webhook output to
    create a Spotify playlist of those tracks.

    Args:
        config: Configuration object.

    Raises:
        RuntimeError: If clipboard is empty.
    """
    data = pyperclip.paste()
    if not data:
        raise RuntimeError(
            "Generating a Spotify playlist from an upload requires output "
            "from an upload_music Discord webhook to be copied to the "
            "system's clipboard"
        )

    spotify = get_spotify_client(config)
    playlist_ids = get_playlist_ids()

    # Parse (track title, artist name) tuples from upload output
    user = ""
    files = []
    for line in data.split("\n"):
        if not line.startswith(" "):
            if not user:
                user = line.split("/")[0]
            continue
        file_ = Path(line).stem
        try:
            track, artist = file_.strip().split(" - ")
        except ValueError:
            logger.warning(f"{line.strip()} is not a valid file")
            continue
        if config.sync.artist_first:
            track, artist = artist, track
        files.append((track, artist))

    files = list(filter(lambda x: len(x) == 2, files))

    # Search Spotify for each file
    threshold = config.spotify.spotify_playlist_fuzz_ratio
    tracks = []
    for title, artist in files:
        query = f"track:{title} artist:{artist}"
        try:
            results = spotify.search(q=query, type="track", limit=50)
        except Exception as exc:
            logger.error(f'Error searching for "{title} - {artist}": {exc}')
            continue

        match, _ = filter_results(spotify, results, threshold, title, artist)
        if match:
            artists = ", ".join([y["name"] for y in match["artists"]])
            logger.info(
                f"Matched {match['name']} - {artists} to {title} - {artist}"
            )
        else:
            logger.warning(f"Could not find a match for {title} - {artist}")
            continue
        tracks.append((match["id"], f'{match["name"]} - {artists}'))

    # Populate playlist
    playlist_ids = populate_playlist(
        playlist_name=f"{user} Uploads",
        playlist_ids=playlist_ids,
        spotify_username=config.spotify.spotify_username,
        spotify=spotify,
        tracks=tracks,
        verbosity=config.verbosity,
    )

    write_playlist_ids(playlist_ids)


def spotify_playlists(config: BaseConfig):
    """Entry point for async Spotify playlist updates.

    Args:
        config: Configuration object.
    """
    asyncio.run(async_spotify_playlists(config))



"""Testing for the check_tracks module."""

from pathlib import Path
from unittest import mock

import pytest

from djtools.utils.check_tracks import compare_tracks


@pytest.mark.parametrize("download_from_spotify", [False, True])
@mock.patch(
    "djtools.utils.check_tracks.get_beatcloud_tracks",
    new=mock.Mock(return_value=[]),
)
@mock.patch(
    "djtools.utils.check_tracks.get_spotify_tracks",
    new=mock.Mock(return_value=[]),
)
@mock.patch("djtools.utils.check_tracks.get_local_tracks")
@mock.patch("djtools.utils.helpers.get_spotify_client", new=mock.Mock())
def test_compare_tracks_ignores_local_dirs_with_download_spotify_playlist(
    mock_get_local_tracks, download_from_spotify, config
):
    """Test the compare_tracks function."""
    if download_from_spotify:
        config.sync.download_spotify_playlist = "test-playlist"
    else:
        config.sync.download_spotify_playlist = ""
    local_dirs = [Path("some/dir")]
    config.utils.local_dirs = local_dirs
    mock_get_local_tracks.return_value = []
    compare_tracks(config)
    if download_from_spotify:
        mock_get_local_tracks.assert_not_called()
    else:
        mock_get_local_tracks.assert_called_once()
    assert config.utils.local_dirs == local_dirs


@mock.patch(
    "djtools.utils.check_tracks.find_matches",
    new=mock.Mock(return_value=[]),
)
@mock.patch(
    "djtools.utils.check_tracks.get_beatcloud_tracks",
    new=mock.Mock(return_value=[]),
)
@mock.patch(
    "djtools.utils.check_tracks.get_spotify_tracks",
    new=mock.Mock(return_value={}),
)
@pytest.mark.parametrize("downloading_instead_of_checking", [True, False])
def test_compare_tracks_get_spotify_tracks_no_tracks(
    downloading_instead_of_checking, config, caplog
):
    """Test the compare_tracks function."""
    caplog.set_level("WARNING")
    if downloading_instead_of_checking:
        config.sync.download_spotify_playlist = "test-playlist"
    else:
        config.utils.check_tracks_spotify_playlists = ["test-playlist"]
    beatcloud_tracks, beatcloud_matches = compare_tracks(config)
    if downloading_instead_of_checking:
        assert (
            "download_spotify_playlist is a key" in caplog.records[0].message
        )
    else:
        assert (
            "check_tracks_spotify_playlists has one or more keys"
            in caplog.records[0].message
        )
    assert not beatcloud_tracks
    assert not beatcloud_matches


@pytest.mark.parametrize("artist_first", [False, True])
@mock.patch("djtools.utils.check_tracks.find_matches")
@mock.patch("djtools.utils.check_tracks.get_beatcloud_tracks")
@mock.patch("djtools.utils.check_tracks.get_spotify_tracks")
def test_compare_tracks_get_spotify_tracks_yields_tracks(
    mock_get_spotify_tracks,
    mock_get_beatcloud_tracks,
    mock_find_matches,
    artist_first,
    config,
    caplog,
):
    """Test the compare_tracks function."""
    caplog.set_level("INFO")
    config.utils.check_tracks_spotify_playlists = ["test-playlist"]
    config.sync.artist_first = artist_first
    title = "title"
    artist = "artist"
    get_spotify_tracks_result = {
        "test-playlist": [
            {"track": {"name": title, "artists": [{"name": artist}]}}
        ]
    }
    expected_match = (
        f"{artist} - {title}" if artist_first else f"{title} - {artist}"
    )
    find_matches_result = [
        ("test-playlist", expected_match, expected_match, 100)
    ]
    expected_matches = [Path(f"{expected_match}.mp3")]
    mock_get_spotify_tracks.return_value = get_spotify_tracks_result
    mock_find_matches.return_value = find_matches_result
    mock_get_beatcloud_tracks.return_value = expected_matches
    beatcloud_tracks, beatcloud_matches = compare_tracks(config)
    assert (
        caplog.records[0].message
        == "\nSpotify Playlist Tracks / Beatcloud Matches: 1"
    )
    assert caplog.records[1].message == "test-playlist:"
    assert (
        caplog.records[2].message
        == f"\t100: {expected_match} | {expected_match}"
    )
    assert beatcloud_tracks == mock_get_beatcloud_tracks.return_value
    assert beatcloud_matches == expected_matches


@mock.patch(
    "djtools.utils.check_tracks.find_matches",
    new=mock.Mock(return_value=[]),
)
@mock.patch(
    "djtools.utils.check_tracks.get_beatcloud_tracks",
    new=mock.Mock(return_value=[]),
)
@mock.patch(
    "djtools.utils.check_tracks.get_local_tracks",
    new=mock.Mock(return_value={}),
)
def test_compare_tracks_get_local_tracks_no_tracks(config, caplog):
    """Test the compare_tracks function."""
    caplog.set_level("WARNING")
    config.utils.local_dirs = [Path("some/dir")]
    beatcloud_tracks, beatcloud_matches = compare_tracks(config)
    assert caplog.records[0].message == (
        "There are no local tracks; make sure local_dirs has one or "
        "more directories containing one or more tracks"
    )
    assert not beatcloud_tracks
    assert not beatcloud_matches


@pytest.mark.parametrize("artist_first", [False, True])
@mock.patch("djtools.utils.check_tracks.find_matches")
@mock.patch("djtools.utils.check_tracks.get_beatcloud_tracks")
@mock.patch("djtools.utils.check_tracks.get_local_tracks")
@mock.patch("djtools.utils.check_tracks.reverse_title_and_artist")
def test_compare_tracks_get_local_tracks_yields_tracks(
    mock_reverse_title_and_artist,
    mock_get_local_tracks,
    mock_get_beatcloud_tracks,
    mock_find_matches,
    artist_first,
    config,
):
    """Test the compare_tracks function."""
    local_dir = Path("some/dir")
    config.utils.local_dirs = [local_dir]
    config.sync.artist_first = artist_first
    file_stem = "artist - title" if artist_first else "title - artist"
    mock_get_local_tracks.return_value = {
        local_dir: [Path(f"{file_stem}.mp3")]
    }
    mock_get_beatcloud_tracks.return_value = [Path(f"{file_stem}.mp3")]
    find_matches_result = [(local_dir, file_stem, file_stem, 100)]
    mock_find_matches.return_value = find_matches_result
    compare_tracks(config)
    if artist_first:
        mock_reverse_title_and_artist.assert_called_once()
    else:
        mock_reverse_title_and_artist.assert_not_called()


@pytest.mark.parametrize(
    "beatcloud_tracks", [[Path("title - artist.mp3")], []]
)
@mock.patch("djtools.utils.check_tracks.get_beatcloud_tracks")
@mock.patch(
    "djtools.utils.check_tracks.get_spotify_tracks",
    new=mock.Mock(
        return_value={
            "test-playlist": [
                {"track": {"name": "title", "artists": [{"name": "artist"}]}}
            ]
        }
    ),
)
@mock.patch(
    "djtools.utils.check_tracks.find_matches",
    new=mock.Mock(return_value=[]),
)
def test_compare_tracks_cached_get_beatcloud_tracks(
    mock_get_beatcloud_tracks, beatcloud_tracks, config
):
    """Test the compare_tracks function."""
    config.utils.check_tracks_spotify_playlists = ["test-playlist"]
    compare_tracks(config, beatcloud_tracks)
    if beatcloud_tracks:
        mock_get_beatcloud_tracks.assert_not_called()
    else:
        mock_get_beatcloud_tracks.return_value = beatcloud_tracks
        mock_get_beatcloud_tracks.assert_called_once()



"""This script is used to populate a Spotify playlist with tracks added to a
collection.
"""

# pylint: disable=missing-function-docstring,inconsistent-return-statements,protected-access,redefined-outer-name,R0801
import logging
from argparse import ArgumentParser, Namespace
from datetime import datetime
from pathlib import Path
from typing import Union

from spotipy import Spotify
from tqdm import tqdm

from djtools.configs.config import BaseConfig
from djtools.collection import RekordboxCollection
from djtools.collection.base_track import Track
from djtools.configs import build_config
from djtools.spotify.helpers import (
    get_spotify_client,
    get_playlist_ids,
    filter_results,
    populate_playlist,
    write_playlist_ids,
)


logger = logging.getLogger(__name__)
logger.setLevel(logging.CRITICAL)
REQ_ID_LIMIT = 1


def parse_args() -> Namespace:
    parser = ArgumentParser()
    parser.add_argument("--config")
    parser.add_argument(
        "--date-filter",
        type=convert_to_datetime,
        default="most-recent",
        help=(
            "Datetime to after which tracks should have tags added from "
            "Spotify. Default is 'most-recent'."
        ),
    )
    parser.add_argument("--playlist-name", default="Library Tracks")
    args = parser.parse_args()

    return args


def convert_to_datetime(arg: str) -> Union[datetime, str]:
    """Convert string to datetime.

    Args:
        arg: datetime as a YYYY-MM-DD string

    Returns:
        Either the datetime object to filter after or a string indicating that
            the most recent upload should be filtered.
    """
    if arg in ["most-recent", "all"]:
        return arg

    try:
        return datetime.strptime(arg, "%Y-%m-%d")
    except ValueError as exc:
        raise ValueError(
            f"Couldn't convert --date-filter argument '{arg}' into a datetime."
        ) from exc


def filter_tracks(
    tracks: dict[str, Track], date_filter: Union[datetime, str]
) -> dict[str, Track]:
    """Filter tracks using a date filter.

    Args:
        tracks: Dictionary of tracks to filter.
        date_filter: datetime to filter tracks with.

    Returns:
        Dictionary of filtered tracks.
    """
    if date_filter == "all":
        return tracks

    sorted_tracks = sorted(
        tracks.values(), key=lambda x: x.get_date_added(), reverse=True
    )

    if date_filter == "most-recent":
        date_filter = sorted_tracks[0].get_date_added()

    filtered_tracks = []
    for track in sorted_tracks:
        if track.get_date_added() < date_filter:
            break
        filtered_tracks.append(track)

    return {track.get_id(): track for track in filtered_tracks}


def find_track(track: Track, config: BaseConfig, spotify: Spotify):
    title = track._Name
    artist = track.get_artists()
    threshold = config.spotify.spotify_playlist_fuzz_ratio
    query = f"track:{title} artist:{artist}"
    try:
        results = spotify.search(q=query, type="track", limit=50)
    except Exception as exc:
        logger.error(f'Error searching for "{title} - {artist}": {exc}')
        return

    match, _ = filter_results(spotify, results, threshold, title, artist)
    if match:
        artists = ", ".join([y["name"] for y in match["artists"]])
        logger.warning(
            f"Matched {match['name']} - {artists} to {title} - {artist}"
        )
    else:
        logger.warning(f"Could not find a match for {title} - {artist}")
        return

    return (match["id"], f'{match["name"]} - {artists}')


def main(config_path: Path, date_filter: datetime, playlist_name: str):
    config = build_config(config_path)
    spotify = get_spotify_client(config)
    playlist_ids = get_playlist_ids()

    collection = RekordboxCollection(config.collection.collection_path)
    tracks = filter_tracks(collection.get_tracks(), date_filter)
    tracks = sorted(tracks.values(), key=lambda x: x.get_date_added())

    found_tracks = list(
        filter(
            None,
            [
                find_track(track, config, spotify)
                for track in tqdm(tracks, desc="Searching Spotify")
            ],
        )
    )

    num_chunks = len(found_tracks) // REQ_ID_LIMIT
    num_chunks += 0 if len(found_tracks) / REQ_ID_LIMIT == 0 else 1
    for index in tqdm(
        range(0, len(found_tracks), REQ_ID_LIMIT), total=num_chunks
    ):
        chunk = found_tracks[index : index + REQ_ID_LIMIT]
        playlist_ids = populate_playlist(
            playlist_name=playlist_name,
            playlist_ids=playlist_ids,
            spotify_username=config.spotify.spotify_username,
            spotify=spotify,
            tracks=chunk,
            verbosity=config.verbosity,
        )

        write_playlist_ids(playlist_ids)


if __name__ == "__main__":
    args = parse_args()
    main(
        args.config,
        args.date_filter,
        args.playlist_name,
    )



"""Testing for the config module."""

import os
import re
from pathlib import Path
from unittest import mock

import getpass
import pytest


from djtools.sync.config import SyncConfig


@pytest.mark.parametrize("download,upload", [(True, False), (False, True)])
def test_syncconfig_download_or_upload_without_usb_path(download, upload):
    """Test for the SyncConfig class."""
    cfg = {
        "aws_profile": "myprofile",
        "bucket_url": "s3://some-bucket.com",
        "download_music": download,
        "upload_music": upload,
        "usb_path": None,
    }
    with pytest.raises(
        RuntimeError,
        match=(
            "Config must include usb_path for both download_music and "
            "upload_music sync operations"
        ),
    ):
        SyncConfig(**cfg)


@pytest.mark.parametrize("download,upload", [(True, False), (False, True)])
def test_syncconfig_download_or_upload_with_missing_usb_path(download, upload):
    """Test for the SyncConfig class."""
    usb_path = Path("not/real/usb/path")
    cfg = {
        "aws_profile": "myprofile",
        "bucket_url": "s3://some-bucket.com",
        "download_music": download,
        "upload_music": upload,
        "usb_path": usb_path,
    }
    with pytest.raises(
        RuntimeError,
        match=re.escape(f'Configured usb_path "{usb_path}" was not found!'),
    ):
        SyncConfig(**cfg)


def test_syncconfig_download_without_import_user():
    """Test for the SyncConfig class."""
    cfg = {
        "download_collection": True,
        "aws_profile": "myprofile",
        "bucket_url": "s3://some-bucket.com",
        "import_user": "",
    }
    with pytest.raises(
        RuntimeError,
        match="import_user must be set to download a collection",
    ):
        SyncConfig(**cfg)


@pytest.mark.parametrize("sync_direction", ["down", "up"])
def test_syncconfig_mutually_exclusive_dirs(sync_direction):
    """Test for the SyncConfig class."""
    cfg = {
        f"{sync_direction}load_include_dirs": ["test"],
        f"{sync_direction}load_exclude_dirs": ["test"],
    }
    with pytest.raises(
        ValueError,
        match=(
            "Config must neither contain both upload_include_dirs and "
            "upload_exclude_dirs or both download_include_dirs and "
            "download_exclude_dirs"
        ),
    ):
        SyncConfig(**cfg)


@pytest.mark.parametrize(
    "aws_operation",
    [
        "download_collection",
        "download_music",
        "upload_collection",
        "upload_music",
    ],
)
def test_syncconfig_no_aws_profile(aws_operation):
    """Test for the SyncConfig class."""
    cfg = {"aws_profile": "", aws_operation: True}
    with pytest.raises(
        RuntimeError,
        match="Config must include aws_profile for sync operations",
    ):
        SyncConfig(**cfg)


@pytest.mark.parametrize(
    "aws_operation",
    [
        "download_collection",
        "download_music",
        "upload_collection",
        "upload_music",
    ],
)
def test_syncconfig_no_bucket(aws_operation):
    """Test for the SyncConfig class."""
    cfg = {"aws_profile": "default", aws_operation: True}
    with pytest.raises(
        RuntimeError,
        match="Config must include bucket_url for sync operations",
    ):
        SyncConfig(**cfg)


def test_syncconfig_sets_aws_profile_env_var():
    """Test for the SyncConfig class."""
    cfg = {"aws_profile": "test-profile"}
    assert (
        os.environ.get("AWS_PROFILE")  # pylint: disable=no-member
        != cfg["aws_profile"]
    )
    SyncConfig(**cfg)
    assert (
        os.environ.get("AWS_PROFILE")  # pylint: disable=no-member
        == cfg["aws_profile"]
    )


@pytest.mark.parametrize(
    "input_user,output_user",
    [("", getpass.getuser()), ("test-user", "test-user")],
)
def test_syncconfig_set_user(input_user, output_user):
    """Test for the SyncConfig class."""
    cfg = {"user": input_user}
    assert not SyncConfig.model_fields.get("user").default
    sync_config = SyncConfig(**cfg)
    assert sync_config.user == output_user


@mock.patch("djtools.spotify.helpers.get_spotify_client", mock.MagicMock())
def test_syncconfig_upload_without_discord_url(caplog):
    """Test for the SyncConfig class."""
    caplog.set_level("WARNING")
    cfg = {
        "bucket_url": "s3://some-bucket.com",
        "discord_url": "",
        "upload_music": True,
        "usb_path": ".",
    }
    SyncConfig(**cfg)
    assert caplog.records[0].message == (
        'discord_url is not configured...set this for "New Music" '
        "discord messages!"
    )


@pytest.mark.parametrize(
    "path", ["some/path/string", Path("some/path/object")]
)
def test_syncconfig_usb_path_is_path(path):
    """Test for the SyncConfig class."""
    cfg = {"usb_path": path}
    sync_config = SyncConfig(**cfg)
    assert isinstance(sync_config.usb_path, Path)
